#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.assertions import (
    assertion_for_account_and_region_task,
)
from servicecatalog_puppet.workflow.assertions import assertion_for_account_task
from servicecatalog_puppet.workflow.assertions import assertion_for_region_task
from servicecatalog_puppet.workflow.assertions import assertion_task

from servicecatalog_puppet.workflow.generic import generic_section_task


class AssertionsSectionTask(generic_section_task.GenericSectionTask):
    section_name_singular = constants.ASSERTION
    section_name = constants.ASSERTIONS
    for_region_task_klass = assertion_for_region_task.AssertionForRegionTask
    for_account_task_klass = assertion_for_account_task.AssertionForAccountTask
    for_account_and_region_task_klass = (
        assertion_for_account_and_region_task.AssertionForAccountAndRegionTask
    )
    task_klass = assertion_task.AssertionTask
    item_name = "assertion_name"
    supports_spoke_mode = True



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.general import get_ssm_param_task
from servicecatalog_puppet.workflow.manifest import manifest_mixin


class ProvisioningTask(
    get_ssm_param_task.PuppetTaskWithParameters, manifest_mixin.ManifestMixen
):
    manifest_file_path = luigi.Parameter()

    @property
    def section_name(self):
        return constants.LAUNCHES

    @property
    def item_name(self):
        return self.launch_name

    @property
    def item_identifier(self):
        return "launch_name"



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import io
import json
import zipfile

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow import dependency
from servicecatalog_puppet.workflow.general import get_ssm_param_task
from servicecatalog_puppet.workflow.manifest import manifest_mixin
from servicecatalog_puppet.workflow.workspaces import Limits
from servicecatalog_puppet.workflow.workspaces import prepare_account_for_workspace_task
from servicecatalog_puppet.workflow.workspaces import workspace_base_task


class ProvisionWorkspaceTask(
    workspace_base_task.WorkspaceBaseTask,
    get_ssm_param_task.PuppetTaskWithParameters,
    manifest_mixin.ManifestMixen,
    dependency.DependenciesMixin,
):
    workspace_name = luigi.Parameter()
    region = luigi.Parameter()
    account_id = luigi.Parameter()

    bucket = luigi.Parameter()
    key = luigi.Parameter()
    version_id = luigi.Parameter()

    puppet_account_id = luigi.Parameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    retry_count = luigi.IntParameter(default=1, significant=False)
    worker_timeout = luigi.IntParameter(default=0, significant=False)
    ssm_param_outputs = luigi.ListParameter(default=[], significant=False)
    requested_priority = luigi.IntParameter(significant=False, default=0)

    execution = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "workspace_name": self.workspace_name,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        requirements = {
            "section_dependencies": self.get_section_dependencies(),
            "ssm_params": self.get_parameters_tasks(),
        }
        if not self.is_running_in_spoke():
            requirements[
                "account_ready"
            ] = prepare_account_for_workspace_task.PrepareAccountForWorkspaceTask(
                puppet_account_id=self.puppet_account_id, account_id=self.account_id,
            )
        return requirements

    def resources_used(self):
        return [
            (self.account_id, Limits.CODEBUILD_CONCURRENT_PROJECTS),
        ]

    def run(self):
        with self.hub_client("s3") as s3:
            options = (
                zipfile.ZipFile(
                    io.BytesIO(
                        s3.get_object(Bucket=self.bucket, Key=self.key)
                        .get("Body")
                        .read()
                    )
                )
                .open(f"options.json", "r")
                .read()
            )

        options = json.loads(options)

        zip_file_path = f"s3://{self.bucket}/{self.key}"
        state_file_path = f"s3://sc-puppet-state-{self.account_id}/workspace/{self.workspace_name}/{self.account_id}/{self.region}.zip"
        with self.spoke_client("codebuild") as codebuild:
            parameters_to_use = [
                dict(name="TARGET_ACCOUNT", value=self.account_id, type="PLAINTEXT",),
                dict(name="STATE_FILE", value=state_file_path, type="PLAINTEXT",),
                dict(name="ZIP", value=zip_file_path, type="PLAINTEXT",),
            ]

            for parameter_name, parameter_value in self.get_parameter_values().items():
                parameters_to_use.append(
                    dict(
                        name=f"TF_VAR_{parameter_name}",
                        value=f"{parameter_value}",
                        type="PLAINTEXT",
                    ),
                )

            parameters_to_use.append(
                dict(
                    name="TERRAFORM_VERSION",
                    value=options.get("Terraform", {}).get(
                        "Version", constants.DEFAULT_TERRAFORM_VERSION_VALUE
                    ),
                    type="PLAINTEXT",
                ),
            )

            build = codebuild.start_build_and_wait_for_completion(
                projectName=constants.EXECUTE_TERRAFORM_PROJECT_NAME,
                environmentVariablesOverride=parameters_to_use,
            )

        if len(self.ssm_param_outputs) > 0:
            with self.spoke_client("s3") as s3:
                output_bucket = f"sc-puppet-state-{self.account_id}"
                output_key = f"terraform-executions/{build.get('id').split(':')[1]}/artifacts-execute/outputs.json"
                outputs = json.loads(
                    s3.get_object(Bucket=output_bucket, Key=output_key)
                    .get("Body")
                    .read()
                )

                for ssm_param_output in self.ssm_param_outputs:
                    self.info(
                        f"writing SSM Param: {ssm_param_output.get('stack_output')}"
                    )
                    with self.hub_client("ssm") as ssm:
                        if outputs.get(ssm_param_output.get("stack_output")):
                            output_value = outputs.get(
                                ssm_param_output.get("stack_output")
                            ).get("value")

                            ssm_parameter_name = ssm_param_output.get("param_name")
                            ssm_parameter_name = ssm_parameter_name.replace(
                                "${AWS::Region}", self.region
                            )
                            ssm_parameter_name = ssm_parameter_name.replace(
                                "${AWS::AccountId}", self.account_id
                            )
                            ssm.put_parameter_and_wait(
                                Name=ssm_parameter_name,
                                Value=output_value,
                                Type=ssm_param_output.get("param_type", "String"),
                                Overwrite=True,
                            )
                        else:
                            raise Exception(
                                f"Could not find {ssm_param_output.get('stack_output')} in the outputs"
                            )

        self.write_output(self.params_for_results_display())



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class GeneratePoliciesTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    puppet_account_id = "puppet_account_id"
    manifest_file_path = "manifest_file_path"
    region = "region"
    sharing_policies = {}

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.generate import generate_policies_task

        self.module = generate_policies_task

        self.sut = self.module.GeneratePolicies(
            puppet_account_id=self.puppet_account_id,
            manifest_file_path=self.manifest_file_path,
            region=self.region,
            sharing_policies=self.sharing_policies,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "manifest_file_path": self.manifest_file_path,
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_api_calls_used(self):
        # setup
        expected_result = {
            f"cloudformation.create_or_update_{self.puppet_account_id}_{self.region}": 1,
        }

        # exercise
        actual_result = self.sut.api_calls_used()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_run(self):
        # setup
        template = "foobar"
        self.inject_into_input("template", template)

        # exercise
        self.sut.run()

        # verify
        self.assert_hub_regional_client_called_with(
            "cloudformation",
            "create_or_update",
            dict(
                ShouldUseChangeSets=False,
                StackName="servicecatalog-puppet-policies",
                TemplateBody=template,
                NotificationARNs=[],
            ),
        )



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import unittest
from unittest import mock


def mocked_client():
    context_handler_mock = mock.MagicMock()
    client_mock = mock.MagicMock()
    context_handler_mock.return_value.__enter__.return_value = client_mock
    return client_mock, context_handler_mock


class FakeInput(object):

    values = None

    def __init__(self) -> None:
        self.values = dict()

    def get_value(self, name):
        return self.values.get(name)

    def set_value(self, name, value):
        self.values[name] = value


class PuppetTaskUnitTest(unittest.TestCase):
    cache_invalidator = "NOW"

    def wire_up_mocks(self):
        self.spoke_client_mock, self.sut.spoke_client = mocked_client()
        (
            self.spoke_regional_client_mock,
            self.sut.spoke_regional_client,
        ) = mocked_client()

        self.hub_client_mock, self.sut.hub_client = mocked_client()
        self.hub_regional_client_mock, self.sut.hub_regional_client = mocked_client()

        self.client_mock, self.sut.client = mocked_client()
        self.regional_client_mock, self.sut.regional_client = mocked_client()

        self.sut.write_output = mock.MagicMock()
        self.sut.input = mock.MagicMock()

        self.fake_inputs = FakeInput()
        self.sut.read_from_input = self.fake_inputs.get_value

    def assert_client_called_with(
        self,
        spoke_regional_client,
        spoke_regional_client_mock,
        client_used,
        function_name_called,
        function_parameters,
        extra_args={},
    ):
        spoke_regional_client.assert_called_once_with(client_used, **extra_args)

        function_called = getattr(spoke_regional_client_mock, function_name_called)
        function_called.assert_called_once_with(**function_parameters)

    def assert_spoke_regional_client_called_with(
        self, client_used, function_name_called, function_parameters
    ):
        self.assert_client_called_with(
            self.sut.spoke_regional_client,
            self.spoke_regional_client_mock,
            client_used,
            function_name_called,
            function_parameters,
        )

    def assert_hub_regional_client_called_with(
        self, client_used, function_name_called, function_parameters, extra_args={}
    ):
        self.assert_client_called_with(
            self.sut.hub_regional_client,
            self.hub_regional_client_mock,
            client_used,
            function_name_called,
            function_parameters,
            extra_args,
        )

    def assert_regional_client_called_with(
        self, client_used, function_name_called, function_parameters, extra_args={}
    ):
        self.assert_client_called_with(
            self.sut.regional_client,
            self.regional_client_mock,
            client_used,
            function_name_called,
            function_parameters,
            extra_args,
        )

    def inject_client_with_response(self, client, function_name, response):
        f = getattr(client, function_name)
        f.return_value = response

    def inject_hub_regional_client_called_with_response(
        self, client_used, function_name_called, response
    ):
        self.inject_client_with_response(
            self.hub_regional_client_mock, function_name_called, response
        )

    def inject_regional_client_called_with_response(
        self, client_used, function_name_called, response
    ):
        self.inject_client_with_response(
            self.regional_client_mock, function_name_called, response
        )

    def inject_spoke_regional_client_called_with_response(
        self, client_used, function_name_called, response
    ):
        self.inject_client_with_response(
            self.spoke_regional_client_mock, function_name_called, response
        )

    def inject_into_input(self, name, value):
        self.fake_inputs.set_value(name, value)

    def assert_output(self, expected_output):
        self.sut.write_output.assert_called_once_with(expected_output)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import click

from servicecatalog_puppet import config
from servicecatalog_puppet.commands.misc import generate_tasks
from servicecatalog_puppet.workflow import tasks as puppet_tasks


def graph_nodes(what):
    nodes = []
    if isinstance(what, puppet_tasks.PuppetTask):
        nodes.append(what.graph_node())
        nodes += graph_nodes(what.requires())
    elif isinstance(what, list):
        for i in what:
            nodes += graph_nodes(i)
    elif isinstance(what, dict):
        for i in what.values():
            nodes += graph_nodes(i)
    else:
        raise Exception(f"unknown {type(what)}")
    return nodes


def graph_lines(task, dependency):
    nodes = []
    if isinstance(dependency, puppet_tasks.PuppetTask):
        nodes.append(f'"{task.node_id}" -> "{dependency.node_id}"')
        # nodes += graph_lines(task, task.requires())
    elif isinstance(dependency, list):
        for i in dependency:
            nodes += graph_lines(task, i)
    elif isinstance(dependency, dict):
        for i in dependency.values():
            nodes += graph_lines(task, i)
    else:
        raise Exception(f"unknown {type(dependency)}")
    return nodes


def graph(f):
    current_account_id = puppet_account_id = config.get_puppet_account_id()
    tasks_to_run = generate_tasks(f, puppet_account_id, current_account_id)
    lines = []
    nodes = []
    for task in tasks_to_run:
        nodes += graph_nodes(task)

        what = task.requires()
        if isinstance(what, puppet_tasks.PuppetTask):
            lines.append(f'"{task.node_id}" -> "{what.node_id}"')
        elif isinstance(what, list):
            for item in what:
                lines += graph_lines(task, item)
        elif isinstance(what, dict):
            for item in what.values():
                lines += graph_lines(task, item)
        else:
            raise Exception(f"unknown {type(what)}")

        # nodes.append(task.graph_node())
        # lines += task.get_graph_lines()
    click.echo("digraph G {\n")
    click.echo("node [shape=record fontname=Arial];")
    for node in nodes:
        click.echo(f"{node};")
    for line in lines:
        click.echo(f'{line} [label="depends on"];')
    click.echo("}")



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet.workflow.portfolio.accessors import (
    search_products_as_admin_task,
)
from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
    portfolio_management_task,
)


class GetProductIdByProductName(portfolio_management_task.PortfolioManagementTask):
    puppet_account_id = luigi.Parameter()
    portfolio = luigi.Parameter()
    product = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "product": self.product,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        return {
            "search_products_as_admin": search_products_as_admin_task.SearchProductsAsAdminTask(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                portfolio=self.portfolio,
                account_id=self.account_id,
                region=self.region,
            ),
        }

    def run(self):
        product_id = None
        response = self.load_from_input("search_products_as_admin")
        for product_view_details in response.get("ProductViewDetails"):
            product_view = product_view_details.get("ProductViewSummary")
            self.info(f"looking at product: {product_view.get('Name')}")
            if product_view.get("Name") == self.product:
                self.info("Found product: {}".format(product_view))
                product_id = product_view.get("ProductId")
        assert product_id is not None, "Did not find product looking for"
        self.write_output(
            {
                "product_name": self.product,
                "product_id": product_id,
                "portfolio_name": self.portfolio,
            }
        )



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import logging
import os
import time

import click
import yaml
from betterboto import client as betterboto_client

from servicecatalog_puppet import config, constants

logger = logging.getLogger(__file__)


def terminate_provisioned_product(prefix, service_catalog, provisioned_product_id):
    logger.info(f"{prefix} :: about to terminate {provisioned_product_id}")
    record_detail = service_catalog.terminate_provisioned_product(
        ProvisionedProductId=provisioned_product_id
    ).get("RecordDetail")
    record_id = record_detail.get("RecordId")

    logger.info(
        f"{prefix} :: waiting for termination of {provisioned_product_id} to complete"
    )
    status = "IN_PROGRESS"
    while status == "IN_PROGRESS":
        record_detail = service_catalog.describe_record(Id=record_id).get(
            "RecordDetail"
        )
        status = record_detail.get("Status")
        logger.info(
            f"{prefix} :: termination of {provisioned_product_id} current status: {status}"
        )
        if status != "IN_PROGRESS":
            break
        else:
            time.sleep(3)

    if status not in ["CREATED", "SUCCEEDED"]:
        logger.info(yaml.safe_dump(record_detail.get("RecordErrors")))
        raise Exception(f"Failed to terminate provisioned product: Status = {status}")


def get_provisioned_product_from_scan(
    service_catalog, provisioned_product_name, logging_prefix
):
    logger.info(f"{logging_prefix}: running get_provisioned_product_from_scan")
    paginator = service_catalog.get_paginator("scan_provisioned_products")
    pages = paginator.paginate(AccessLevelFilter={"Key": "Account", "Value": "self"},)

    for page in pages:
        logger.info(
            f"{logging_prefix}: running get_provisioned_product_from_scan paging"
        )
        for provisioned_product in page.get("ProvisionedProducts", []):
            if provisioned_product.get("Name") == provisioned_product_name:
                return provisioned_product

    return None


def terminate_if_status_is_not_available(
    service_catalog, provisioned_product_name, product_id, account_id, region
):
    prefix = f"[{provisioned_product_name}] {account_id}:{region}"
    logger.info(f"{prefix} :: checking if should be terminated")
    provisioned_product = get_provisioned_product_from_scan(
        service_catalog, provisioned_product_name, prefix
    )

    provisioned_product_id = False
    provisioning_artifact_id = False
    current_status = False
    if provisioned_product is not None:
        current_status = provisioned_product.get("Status")
        logger.info(f"{prefix} :: current status is {current_status}")
        if current_status in ["AVAILABLE", "TAINTED"]:
            provisioned_product_id = provisioned_product.get("Id")
            provisioning_artifact_id = provisioned_product.get("ProvisioningArtifactId")
        elif current_status in ["UNDER_CHANGE", "PLAN_IN_PROGRESS"]:
            while True:
                status = (
                    service_catalog.describe_provisioned_product(
                        Id=provisioned_product.get("Id")
                    )
                    .get("ProvisionedProductDetail")
                    .get("Status")
                )
                logger.info(f"{prefix} :: waiting to complete: {status}")
                time.sleep(5)
                if status not in ["UNDER_CHANGE", "PLAN_IN_PROGRESS"]:
                    return terminate_if_status_is_not_available(
                        service_catalog,
                        provisioned_product_name,
                        product_id,
                        account_id,
                        region,
                    )

        elif current_status == "ERROR":
            logger.info(
                f"{prefix} :: terminating as its status is {provisioned_product.get('Status')}"
            )
            terminate_provisioned_product(
                prefix, service_catalog, provisioned_product.get("Id")
            )
    logger.info(f"{prefix} :: Finished terminate_if_status_is_not_available")
    return provisioned_product_id, provisioning_artifact_id, current_status


def get_stack_output_for(cloudformation, stack_name):
    logger.info(f"Getting stack output for {stack_name}")
    return cloudformation.describe_stacks(StackName=stack_name).get("Stacks")[0]


def get_default_parameters_for_stack(cloudformation, stack_name):
    logger.info(f"Getting default parameters for for {stack_name}")
    existing_stack_params_dict = {}
    summary_response = cloudformation.get_template_summary(StackName=stack_name,)
    for parameter in summary_response.get("Parameters"):
        existing_stack_params_dict[parameter.get("ParameterKey")] = parameter.get(
            "DefaultValue"
        )
    return existing_stack_params_dict


def get_parameters_for_stack(cloudformation, stack_name):
    existing_stack_params_dict = get_default_parameters_for_stack(
        cloudformation, stack_name
    )

    stack = get_stack_output_for(cloudformation, stack_name)
    for stack_param in stack.get("Parameters", []):
        existing_stack_params_dict[stack_param.get("ParameterKey")] = stack_param.get(
            "ParameterValue"
        )
    return existing_stack_params_dict


def provision_product_with_plan(
    service_catalog,
    launch_name,
    account_id,
    region,
    product_id,
    provisioning_artifact_id,
    puppet_account_id,
    path_id,
    params,
    version,
    should_use_sns,
):
    uid = f"[{launch_name}] {account_id}:{region}]"
    stack_name = "-".join([constants.PREFIX, account_id, region, launch_name])
    logger.info(f"{uid} :: Checking for an existing plan")
    if puppet_account_id == account_id:
        provisioned_product_plans = service_catalog.list_provisioned_product_plans_single_page().get(
            "ProvisionedProductPlans", []
        )
    else:
        provisioned_product_plans = service_catalog.list_provisioned_product_plans_single_page(
            AccessLevelFilter={"Key": "Account", "Value": "self"}
        ).get(
            "ProvisionedProductPlans", []
        )
    for provisioned_product_plan in provisioned_product_plans:
        logger.info(
            f"{uid} :: Found plan for {provisioned_product_plan.get('ProvisionProductName')}"
        )
        if provisioned_product_plan.get("ProvisionProductName") == launch_name:
            f"{uid} :: Found existing plan, going to terminate it"
            service_catalog.delete_provisioned_product_plan(
                PlanId=provisioned_product_plan.get("PlanId")
            )

    logger.info(f"{uid} :: Creating a plan")
    partition = config.get_partition()
    regional_sns_topic = f"arn:{partition}:sns:{region}:{puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
    provisioning_parameters = []
    for p in params.keys():
        provisioning_parameters.append(
            {"Key": p, "Value": params.get(p),}
        )
    response = service_catalog.create_provisioned_product_plan(
        PlanName=stack_name,
        PlanType="CLOUDFORMATION",
        PathId=path_id,
        ProductId=product_id,
        ProvisionedProductName=launch_name,
        ProvisioningArtifactId=provisioning_artifact_id,
        ProvisioningParameters=provisioning_parameters,
        Tags=[
            {"Key": "ServiceCatalogPuppet:Actor", "Value": "Generated",},
            {"Key": "launch_name", "Value": launch_name,},
            {"Key": "version", "Value": version,},
        ],
        NotificationArns=[regional_sns_topic,] if should_use_sns else [],
    )
    logger.info(f"{uid} :: Plan created, waiting for completion")

    plan_id = response.get("PlanId")
    plan_status = "CREATE_IN_PROGRESS"

    while plan_status == "CREATE_IN_PROGRESS":
        describe_provisioned_product_plan_response = service_catalog.describe_provisioned_product_plan(
            PlanId=plan_id
        )
        plan_status = describe_provisioned_product_plan_response.get(
            "ProvisionedProductPlanDetails"
        ).get("Status")
        logger.info(f"{uid} :: Waiting for product plan: {plan_status}")
        time.sleep(5)

    if plan_status in ["CREATE_SUCCESS", "EXECUTE_SUCCESS"]:
        logger.info(
            f"{uid} :: Plan created, "
            f"changes: {yaml.safe_dump(describe_provisioned_product_plan_response.get('ResourceChanges'))}"
        )

        logger.info(f"{uid} :: executing changes")
        service_catalog.execute_provisioned_product_plan(PlanId=plan_id)

        plan_execute_status = "EXECUTE_IN_PROGRESS"
        while plan_execute_status == "EXECUTE_IN_PROGRESS":
            response = service_catalog.describe_provisioned_product_plan(PlanId=plan_id)
            logger.info(f"{uid} :: executing changes for plan: {plan_id}")
            plan_execute_status = response.get("ProvisionedProductPlanDetails").get(
                "Status"
            )
            logger.info(
                f"{uid} :: waiting for execution to complete: {plan_execute_status}"
            )
            time.sleep(5)

        if plan_execute_status in ["CREATE_SUCCESS", "EXECUTE_SUCCESS"]:
            provisioned_product_id = response.get("ProvisionedProductPlanDetails").get(
                "ProvisionProductId"
            )

            logger.info(f"{uid} :: waiting for change to complete")
            while True:
                response = service_catalog.describe_provisioned_product(
                    Id=provisioned_product_id
                )
                logger.info(
                    f"{uid} :: "
                    f"waiting for provision to complete: {response.get('ProvisionedProductDetail').get('Status')}"
                )
                provisioned_product_detail = response.get("ProvisionedProductDetail")
                execute_status = provisioned_product_detail.get("Status")
                if execute_status in ["AVAILABLE", "EXECUTE_SUCCESS"]:
                    break
                elif execute_status == "TAINTED":
                    service_catalog.delete_provisioned_product_plan(PlanId=plan_id)
                    raise Exception(
                        f"{uid} :: Execute failed: {execute_status}: {provisioned_product_detail.get('StatusMessage')}"
                    )
                elif execute_status == "ERROR":
                    raise Exception(
                        f"{uid} :: Execute failed: {execute_status}: {provisioned_product_detail.get('StatusMessage')}"
                    )
                else:
                    time.sleep(5)

            service_catalog.delete_provisioned_product_plan(PlanId=plan_id)
            return provisioned_product_id

        else:
            raise Exception(f"{uid} :: Plan execute failed: {plan_execute_status}")

    else:
        if plan_status == "CREATE_FAILED" and describe_provisioned_product_plan_response.get(
            "ProvisionedProductPlanDetails"
        ).get(
            "StatusMessage"
        ) in [
            "No updates are to be performed.",
            "The submitted information didn't contain changes. Submit different information to create a change set.",
        ]:
            logger.warn(
                f"{uid} :: Swallowing that plan {plan_status} due to {describe_provisioned_product_plan_response.get('ProvisionedProductPlanDetails').get('StatusMessage')}"
            )
            if (
                describe_provisioned_product_plan_response.get(
                    "ProvisionedProductPlanDetails", {}
                ).get("ProvisionProductId", None)
                is None
            ):
                raise Exception(
                    f"The plan for {uid} resulted in no changes and there was no previously provisioned product"
                )
            return describe_provisioned_product_plan_response.get(
                "ProvisionedProductPlanDetails", {}
            ).get("ProvisionProductId")
        else:
            raise Exception(
                f"{uid} :: Plan failed ({plan_status}): {describe_provisioned_product_plan_response.get('ProvisionedProductPlanDetails').get('StatusMessage')}"
            )


def provision_product(
    service_catalog,
    launch_name,
    account_id,
    region,
    product_id,
    provisioning_artifact_id,
    puppet_account_id,
    path_name,
    params,
    version,
    should_use_sns,
    execution,
):
    partition = config.get_partition()
    uid = f"[{launch_name}] {account_id}:{region}]"
    provisioning_parameters = []
    for p in params.keys():
        if params.get(p) is None:
            raise Exception(
                f"Could not provision {launch_name} in {region} of {account_id}, parameter {p} was None"
            )
        provisioning_parameters.append(
            {"Key": p, "Value": params.get(p),}
        )
    provisioned_product_id = (
        service_catalog.provision_product(
            ProductId=product_id,
            ProvisioningArtifactId=provisioning_artifact_id,
            PathName=path_name,
            ProvisionedProductName=launch_name,
            ProvisioningParameters=provisioning_parameters,
            Tags=[
                {"Key": "ServiceCatalogPuppet:Actor", "Value": "Generated",},
                {"Key": "launch_name", "Value": launch_name,},
                {"Key": "version", "Value": version,},
            ],
            NotificationArns=[
                f"arn:{partition}:sns:{region}:{puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events",
            ]
            if should_use_sns
            else [],
        )
        .get("RecordDetail")
        .get("ProvisionedProductId")
    )
    logger.info(f"{uid}: provisioning started: {provisioned_product_id}")

    if execution == constants.EXECUTION_MODE_ASYNC:
        return provisioned_product_id
    else:
        while True:
            response = service_catalog.describe_provisioned_product(
                Id=provisioned_product_id
            )
            logger.info(
                f"{uid} :: "
                f"waiting for provision to complete: {response.get('ProvisionedProductDetail').get('Status')}"
            )
            provisioned_product_detail = response.get("ProvisionedProductDetail")
            execute_status = provisioned_product_detail.get("Status")
            if execute_status in ["AVAILABLE", "EXECUTE_SUCCESS"]:
                break
            elif execute_status in ["ERROR", "TAINTED"]:
                raise Exception(
                    f"{uid} :: Execute failed: {execute_status}: {provisioned_product_detail.get('StatusMessage')}"
                )
            else:
                time.sleep(5)
        return provisioned_product_id


def update_provisioned_product(
    service_catalog,
    launch_name,
    account_id,
    region,
    product_id,
    provisioning_artifact_id,
    puppet_account_id,
    path_name,
    params,
    version,
    execution,
):
    uid = f"[{launch_name}] {account_id}:{region}]"
    provisioning_parameters = []
    for p in params.keys():
        provisioning_parameters.append(
            {"Key": p, "Value": params.get(p),}
        )
    provisioned_product_id = (
        service_catalog.update_provisioned_product(
            ProductId=product_id,
            ProvisioningArtifactId=provisioning_artifact_id,
            PathName=path_name,
            ProvisionedProductName=launch_name,
            ProvisioningParameters=provisioning_parameters,
        )
        .get("RecordDetail")
        .get("ProvisionedProductId")
    )
    logger.info(f"{uid}: provisioning started: {provisioned_product_id}")

    if execution == constants.EXECUTION_MODE_ASYNC:
        return provisioned_product_id
    else:
        while True:
            response = service_catalog.describe_provisioned_product(
                Id=provisioned_product_id
            )
            logger.info(
                f"{uid} :: "
                f"waiting for provision to complete: {response.get('ProvisionedProductDetail').get('Status')}"
            )
            provisioned_product_detail = response.get("ProvisionedProductDetail")
            execute_status = provisioned_product_detail.get("Status")
            if execute_status in ["AVAILABLE", "EXECUTE_SUCCESS"]:
                break
            elif execute_status in ["ERROR", "TAINTED"]:
                raise Exception(
                    f"{uid} :: Execute failed: {execute_status}: {provisioned_product_detail.get('StatusMessage')}"
                )
            else:
                time.sleep(5)
        return provisioned_product_id


def get_path_for_product(service_catalog, product_id, portfolio_name):
    logger.info(f"Getting path for product {product_id}")
    response = service_catalog.list_launch_paths(ProductId=product_id)
    if len(response.get("LaunchPathSummaries")) == 1:
        path_id = response.get("LaunchPathSummaries")[0].get("Id")
        logger.info(f"There is only one path: {path_id} for product: {product_id}")
        return path_id
    else:
        for launch_path_summary in response.get("LaunchPathSummaries", []):
            name = launch_path_summary.get("Name")
            if name == portfolio_name:
                path_id = launch_path_summary.get("Id")
                logger.info(f"Got path: {path_id} for product: {product_id}")
                return path_id

    raise Exception("Failed to find a LaunchPath")


def ensure_is_terminated(service_catalog, provisioned_product_name, product_id):
    logger.info(f"Ensuring {provisioned_product_name} is terminated")
    r = get_provisioned_product_details(
        product_id, provisioned_product_name, service_catalog
    )

    if r is not None:
        provisioned_product_id = r.get("Id")
        provisioning_artifact_id = r.get("ProvisioningArtifactId")

        if r.get("Status") != "TERMINATED":
            logger.info(
                f"Terminating {provisioned_product_name}, its status is: {r.get('Status')}"
            )
            terminate_provisioned_product(
                f"{provisioned_product_name}:{product_id}", service_catalog, r.get("Id")
            )
        else:
            logger.info(f"Skipping terminated launch: {provisioned_product_name}")

        logger.info(f"Finished ensuring {provisioned_product_name} is terminated")
        return provisioned_product_id, provisioning_artifact_id
    else:
        return None, None


def get_provisioned_product_details(
    product_id, provisioned_product_name, service_catalog
):
    response = service_catalog.scan_provisioned_products_single_page(
        AccessLevelFilter={"Key": "Account", "Value": "self"},
    )
    for r in response.get("ProvisionedProducts", []):
        if r.get("Name") == provisioned_product_name:
            return r
    return None


def ensure_portfolio(service_catalog, portfolio_name, provider_name, description=None):
    return find_portfolio(service_catalog, portfolio_name) or create_portfolio(
        service_catalog, portfolio_name, provider_name, description
    )


def find_portfolio(service_catalog, portfolio_searching_for):
    logger.info("Searching for portfolio for: {}".format(portfolio_searching_for))
    response = service_catalog.list_portfolios_single_page()
    for detail in response.get("PortfolioDetails"):
        if detail.get("DisplayName") == portfolio_searching_for:
            logger.info("Found portfolio: {}".format(portfolio_searching_for))
            return detail
    return False


def create_portfolio(service_catalog, portfolio_name, provider_name, description=None):
    logger.info(f"Creating portfolio: {portfolio_name}")
    args = {
        "DisplayName": portfolio_name,
        "ProviderName": provider_name,
    }
    if description is not None:
        args["Description"] = description
    return service_catalog.create_portfolio(**args).get("PortfolioDetail")


def run_pipeline(pipeline_name, tail):
    with betterboto_client.ClientContextManager("codepipeline") as codepipeline:
        pipeline_execution_id = codepipeline.start_pipeline_execution(
            name=pipeline_name
        ).get("pipelineExecutionId")
        click.echo(
            f"https://{os.environ.get('AWS_DEFAULT_REGION')}.console.aws.amazon.com/codesuite/codepipeline/pipelines/{pipeline_name}/executions/{pipeline_execution_id}/timeline"
        )
        if tail:
            while True:
                pipeline_execution = codepipeline.get_pipeline_execution(
                    pipelineName=pipeline_name,
                    pipelineExecutionId=pipeline_execution_id,
                ).get("pipelineExecution")
                status = pipeline_execution.get("status")
                click.echo(f"status: {status}")
                if status != "InProgress":
                    break
                else:
                    time.sleep(5)
        return pipeline_execution_id



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.general import get_ssm_param_task


class CodeBuildRunBaseTask(get_ssm_param_task.PuppetTaskWithParameters):
    manifest_file_path = luigi.Parameter()

    @property
    def section_name(self):
        return constants.CODE_BUILD_RUNS

    @property
    def item_name(self):
        return self.code_build_run_name

    @property
    def item_identifier(self):
        return "code_build_run_name"



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class DeleteCloudFormationStackTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    account_id = "account_id"
    region = "region"
    stack_name = "stack_name"
    nonce = "nonce"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.general import (
            delete_cloud_formation_stack_task,
        )

        self.module = delete_cloud_formation_stack_task

        self.sut = self.module.DeleteCloudFormationStackTask(
            account_id=self.account_id,
            region=self.region,
            stack_name=self.stack_name,
            nonce=self.nonce,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "stack_name": self.stack_name,
            "account_id": self.account_id,
            "region": self.region,
            "nonce": self.nonce,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_api_calls_used(self):
        # setup
        expected_result = {
            f"cloudformation.describe_stacks_single_page_{self.account_id}_{self.region}": 1,
            f"cloudformation.delete_stack_{self.account_id}_{self.region}": 1,
            f"cloudformation.describe_stack_events_{self.account_id}_{self.region}": 1,
        }

        # exercise
        actual_result = self.sut.api_calls_used()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_run(self):
        # setup
        # exercise
        self.sut.run()
        # verify
        self.assert_spoke_regional_client_called_with(
            "cloudformation", "ensure_deleted", dict(StackName=self.stack_name)
        )
        self.assert_output(self.sut.params_for_results_display())



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.apps import app_for_account_and_region_task
from servicecatalog_puppet.workflow.apps import app_for_account_task
from servicecatalog_puppet.workflow.apps import app_for_region_task
from servicecatalog_puppet.workflow.apps import app_task
from servicecatalog_puppet.workflow.generic import generic_section_task


class AppSectionTask(generic_section_task.GenericSectionTask):
    section_name_singular = constants.APP
    section_name = constants.APPS
    for_region_task_klass = app_for_region_task.AppForRegionTask
    for_account_task_klass = app_for_account_task.AppForAccountTask
    for_account_and_region_task_klass = (
        app_for_account_and_region_task.AppForAccountAndRegionTask
    )
    task_klass = app_task.AppTask
    item_name = "app_name"
    supports_spoke_mode = True



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from unittest import skip

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class BootstrapSpokeAsTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    puppet_account_id = "puppet_account_id"
    account_id = "account_id"
    iam_role_arns = []
    role_name = "role_name"
    permission_boundary = "permission_boundary"
    puppet_role_name = "puppet_role_name"
    puppet_role_path = "puppet_role_path"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow import management

        self.module = management

        self.sut = self.module.BootstrapSpokeAsTask(
            puppet_account_id=self.puppet_account_id,
            account_id=self.account_id,
            iam_role_arns=self.iam_role_arns,
            role_name=self.role_name,
            permission_boundary=self.permission_boundary,
            puppet_role_name=self.puppet_role_name,
            puppet_role_path=self.puppet_role_path,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "account_id": self.account_id,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    @skip
    def test_run(self):
        # setup
        # exercise
        actual_result = self.sut.run()

        # verify
        raise NotImplementedError()



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from unittest import skip

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class CreateLaunchRoleConstraintsForSpokeLocalPortfolioTaskTest(
    tasks_unit_tests_helper.PuppetTaskUnitTest
):
    manifest_file_path = "manifest_file_path"
    spoke_local_portfolio_name = "spoke_local_portfolio_name"
    account_id = "account_id"
    region = "region"
    portfolio = "portfolio"
    puppet_account_id = "puppet_account_id"
    organization = "organization"
    product_generation_method = "product_generation_method"
    launch_constraints = {}
    sharing_mode = "sharing_mode"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.portfolio.constraints_management import (
            create_launch_role_constraints_for_spoke_local_portfolio_task,
        )

        self.module = create_launch_role_constraints_for_spoke_local_portfolio_task

        self.sut = self.module.CreateLaunchRoleConstraintsForSpokeLocalPortfolioTask(
            manifest_file_path=self.manifest_file_path,
            spoke_local_portfolio_name=self.spoke_local_portfolio_name,
            account_id=self.account_id,
            region=self.region,
            portfolio=self.portfolio,
            puppet_account_id=self.puppet_account_id,
            organization=self.organization,
            product_generation_method=self.product_generation_method,
            launch_constraints=self.launch_constraints,
            sharing_mode=self.sharing_mode,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "spoke_local_portfolio_name": self.spoke_local_portfolio_name,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_api_calls_used(self):
        # setup
        expected_result = [
            f"cloudformation.ensure_deleted_{self.account_id}_{self.region}",
            f"cloudformation.describe_stacks_{self.account_id}_{self.region}",
            f"cloudformation.create_or_update_{self.account_id}_{self.region}",
            f"service_catalog.search_products_as_admin_{self.account_id}_{self.region}",
        ]

        # exercise
        actual_result = self.sut.api_calls_used()

        # verify
        self.assertEqual(expected_result, actual_result)

    @skip
    def test_run(self):
        # setup
        # exercise
        actual_result = self.sut.run()

        # verify
        raise NotImplementedError()



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from unittest import skip

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class DoExecuteCodeBuildRunTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    manifest_file_path = "manifest_file_path"
    code_build_run_name = "code_build_run_name"
    puppet_account_id = "puppet_account_id"
    region = "region"
    account_id = "account_id"
    ssm_param_inputs = []
    launch_parameters = {}
    manifest_parameters = {}
    account_parameters = {}
    project_name = "project_name"
    execution = constants.EXECUTION_MODE_HUB
    requested_priority = 1

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.codebuild_runs import (
            do_execute_code_build_run_task,
        )

        self.module = do_execute_code_build_run_task

        self.sut = self.module.DoExecuteCodeBuildRunTask(
            manifest_file_path=self.manifest_file_path,
            code_build_run_name=self.code_build_run_name,
            puppet_account_id=self.puppet_account_id,
            region=self.region,
            account_id=self.account_id,
            ssm_param_inputs=self.ssm_param_inputs,
            launch_parameters=self.launch_parameters,
            manifest_parameters=self.manifest_parameters,
            account_parameters=self.account_parameters,
            project_name=self.project_name,
            requested_priority=self.requested_priority,
            execution=self.execution,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_api_calls_used(self):
        # setup
        expected_result = [
            f"codebuild.start_build_{self.puppet_account_id}_{self.project_name}",
            f"codebuild.batch_get_projects_{self.puppet_account_id}_{self.project_name}",
        ]

        # exercise
        actual_result = self.sut.api_calls_used()

        # verify
        self.assertEqual(expected_result, actual_result)

    @skip
    def test_run(self):
        # setup
        # exercise
        actual_result = self.sut.run()

        # verify
        raise NotImplementedError()



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import json

import luigi

from servicecatalog_puppet import aws
from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.launch import provisioning_task


class TerminateProductDryRunTask(provisioning_task.ProvisioningTask):
    launch_name = luigi.Parameter()
    portfolio = luigi.Parameter()
    portfolio_id = luigi.Parameter()
    product = luigi.Parameter()
    product_id = luigi.Parameter()
    version = luigi.Parameter()
    version_id = luigi.Parameter()

    account_id = luigi.Parameter()
    region = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    retry_count = luigi.IntParameter(default=1)

    ssm_param_outputs = luigi.ListParameter(default=[])

    worker_timeout = luigi.IntParameter(default=0, significant=False)

    parameters = luigi.ListParameter(default=[])
    ssm_param_inputs = luigi.ListParameter(default=[])

    try_count = 1

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "launch_name": self.launch_name,
            "account_id": self.account_id,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
        }

    def write_result(self, current_version, new_version, effect, notes=""):
        with self.output().open("w") as f:
            f.write(
                json.dumps(
                    {
                        "current_version": current_version,
                        "new_version": new_version,
                        "effect": effect,
                        "notes": notes,
                        "params": self.param_kwargs,
                    },
                    indent=4,
                    default=str,
                )
            )

    def api_calls_used(self):
        return [
            f"servicecatalog.scan_provisioned_products_single_page_{self.account_id}_{self.region}",
            f"servicecatalog.describe_provisioning_artifact_{self.account_id}_{self.region}",
        ]

    def run(self):
        self.info(
            f"starting dry run terminate try {self.try_count} of {self.retry_count}"
        )

        with self.spoke_regional_client("servicecatalog") as service_catalog:
            self.info(
                f"[{self.launch_name}] {self.account_id}:{self.region} :: looking for previous failures"
            )
            r = aws.get_provisioned_product_details(
                self.product_id, self.launch_name, service_catalog
            )

            if r is None:
                self.write_result(
                    "-", "-", constants.NO_CHANGE, notes="There is nothing to terminate"
                )
            else:
                provisioned_product_name = (
                    service_catalog.describe_provisioning_artifact(
                        ProvisioningArtifactId=r.get("ProvisioningArtifactId"),
                        ProductId=self.product_id,
                    )
                    .get("ProvisioningArtifactDetail")
                    .get("Name")
                )

                if r.get("Status") != "TERMINATED":
                    self.write_result(
                        provisioned_product_name,
                        "-",
                        constants.CHANGE,
                        notes="The product would be terminated",
                    )
                else:
                    self.write_result(
                        "-",
                        "-",
                        constants.CHANGE,
                        notes="The product is already terminated",
                    )



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import json

import luigi

from servicecatalog_puppet import aws
from servicecatalog_puppet.workflow.manifest import manifest_mixin
from servicecatalog_puppet.workflow.portfolio.accessors import (
    get_portfolio_by_portfolio_name_task,
)
from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
    portfolio_management_task,
)
from servicecatalog_puppet.workflow.portfolio.sharing_management import (
    create_share_for_account_launch_region_task,
)


class CreateSpokeLocalPortfolioTask(
    portfolio_management_task.PortfolioManagementTask, manifest_mixin.ManifestMixen
):
    puppet_account_id = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    organization = luigi.Parameter(significant=False)
    sharing_mode = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        return {
            "create_share_for_account_launch_region": create_share_for_account_launch_region_task.CreateShareForAccountLaunchRegion(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                portfolio=self.portfolio,
                account_id=self.account_id,
                region=self.region,
                sharing_mode=self.sharing_mode,
            ),
            "puppet_portfolio": get_portfolio_by_portfolio_name_task.GetPortfolioByPortfolioName(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                portfolio=self.portfolio,
                account_id=self.puppet_account_id,
                region=self.region,
            ),
        }

    def api_calls_used(self):
        return [
            f"servicecatalog.list_portfolios_{self.account_id}_{self.region}",
            f"servicecatalog.create_portfolio_{self.account_id}_{self.region}",
        ]

    def run(self):
        with self.input().get("puppet_portfolio").open("r") as f:
            portfolio_details = json.loads(f.read())
        with self.spoke_regional_client("servicecatalog") as spoke_service_catalog:
            spoke_portfolio = aws.ensure_portfolio(
                spoke_service_catalog,
                self.portfolio,
                portfolio_details.get("provider_name"),
                portfolio_details.get("description"),
            )
        self.write_output(spoke_portfolio)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi


class GenericForAccountAndRegionTask:
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "account_id": self.account_id,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
            self.item_identifier: self.item_name,
        }

    def requires(self):
        dependencies = list()
        these_dependencies = list()
        requirements = dict(
            dependencies=dependencies, these_dependencies=these_dependencies,
        )

        klass = self.get_klass_for_provisioning()

        a = {
            "puppet_account_id": self.puppet_account_id,
            "section_name": self.section_name,
            self.item_identifier: self.item_name,
            "account_id": self.account_id,
            "region": self.region,
            "single_account": self.single_account,
        }
        for task in self.manifest.get_tasks_for_launch_and_account_and_region(**a):
            dependencies.append(
                klass(**task, manifest_file_path=self.manifest_file_path)
            )

        return requirements



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow import tasks


class EnsureEventBridgeEventBusTask(tasks.PuppetTask):
    puppet_account_id = luigi.Parameter()
    region = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
        }

    def api_calls_used(self):
        return {
            f"events.describe_event_bus_{self.puppet_account_id}_{self.region}": 1,
            f"events.create_event_bus_{self.puppet_account_id}_{self.region}": 1,
        }

    def run(self):
        with self.hub_regional_client("events") as events:
            created = False
            try:
                events.describe_event_bus(Name=constants.EVENT_BUS_NAME)
                created = True
            except events.exceptions.ResourceNotFoundException:
                events.create_event_bus(Name=constants.EVENT_BUS_NAME)
        self.write_output(created)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet.workflow.codebuild_runs import code_build_run_for_task
from servicecatalog_puppet.workflow.generic import generic_for_account_and_region_task


class CodeBuildRunForAccountAndRegionTask(
    generic_for_account_and_region_task.GenericForAccountAndRegionTask,
    code_build_run_for_task.CodeBuildRunForTask,
):
    account_id = luigi.Parameter()
    region = luigi.Parameter()



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import logging

import click
from betterboto import client as betterboto_client
from jinja2 import Template

from servicecatalog_puppet import constants, asset_helpers

logger = logging.getLogger()
logger.setLevel(logging.INFO)


def set_org_iam_role_arn(org_iam_role_arn):
    with betterboto_client.ClientContextManager("ssm") as ssm:
        ssm.put_parameter(
            Name=constants.CONFIG_PARAM_NAME_ORG_IAM_ROLE_ARN,
            Type="String",
            Value=org_iam_role_arn,
            Overwrite=True,
        )
    click.echo("Uploaded config")


def bootstrap_org_master(puppet_account_id):
    with betterboto_client.ClientContextManager("cloudformation",) as cloudformation:
        org_iam_role_arn = None
        logger.info("Starting bootstrap of org master")
        stack_name = f"{constants.BOOTSTRAP_STACK_NAME}-org-master-{puppet_account_id}"
        template = asset_helpers.read_from_site_packages(
            f"{constants.BOOTSTRAP_STACK_NAME}-org-master.template.yaml"
        )
        template = Template(template).render(
            VERSION=constants.VERSION, puppet_account_id=puppet_account_id
        )
        args = {
            "StackName": stack_name,
            "TemplateBody": template,
            "Capabilities": ["CAPABILITY_NAMED_IAM"],
            "Parameters": [
                {
                    "ParameterKey": "PuppetAccountId",
                    "ParameterValue": str(puppet_account_id),
                },
                {
                    "ParameterKey": "Version",
                    "ParameterValue": constants.VERSION,
                    "UsePreviousValue": False,
                },
            ],
            "Tags": [{"Key": "ServiceCatalogPuppet:Actor", "Value": "Framework",}],
        }
        cloudformation.create_or_update(**args)
        response = cloudformation.describe_stacks(StackName=stack_name)
        if len(response.get("Stacks")) != 1:
            raise Exception("Expected there to be only one {} stack".format(stack_name))
        stack = response.get("Stacks")[0]

        for output in stack.get("Outputs"):
            if output.get("OutputKey") == constants.PUPPET_ORG_ROLE_FOR_EXPANDS_ARN:
                logger.info("Finished bootstrap of org-master")
                org_iam_role_arn = output.get("OutputValue")

        if org_iam_role_arn is None:
            raise Exception(
                "Could not find output: {} in stack: {}".format(
                    constants.PUPPET_ORG_ROLE_FOR_EXPANDS_ARN, stack_name
                )
            )

    click.echo("Bootstrapped org master, org-iam-role-arn: {}".format(org_iam_role_arn))



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet.workflow import tasks


class PortfolioManagementTask(tasks.PuppetTask):
    manifest_file_path = luigi.Parameter()



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet.workflow import dependency
from servicecatalog_puppet.workflow.codebuild_runs import code_build_run_base_task
from servicecatalog_puppet.workflow.manifest import manifest_mixin


class DoExecuteCodeBuildRunTask(
    code_build_run_base_task.CodeBuildRunBaseTask,
    manifest_mixin.ManifestMixen,
    dependency.DependenciesMixin,
):
    code_build_run_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    execution = luigi.Parameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    project_name = luigi.Parameter()
    requested_priority = luigi.IntParameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def api_calls_used(self):
        return [
            f"codebuild.start_build_{self.get_account_used()}_{self.project_name}",
            f"codebuild.batch_get_projects_{self.get_account_used()}_{self.project_name}",
        ]

    def requires(self):
        requirements = {
            "ssm_params": self.get_parameters_tasks(),
        }
        return requirements

    def run(self):
        with self.hub_client("codebuild") as codebuild:
            provided_parameters = self.get_parameter_values()
            parameters_to_use = list()

            projects = codebuild.batch_get_projects(names=[self.project_name]).get(
                "projects", []
            )
            for project in projects:
                if project.get("name") == self.project_name:
                    for environment_variable in project.get("environment", {}).get(
                        "environmentVariables", []
                    ):
                        if environment_variable.get("type") == "PLAINTEXT":
                            n = environment_variable.get("name")
                            if provided_parameters.get(n):
                                parameters_to_use.append(
                                    dict(
                                        name=n,
                                        value=provided_parameters.get(n),
                                        type="PLAINTEXT",
                                    )
                                )

            parameters_to_use.append(
                dict(name="TARGET_ACCOUNT_ID", value=self.account_id, type="PLAINTEXT",)
            )
            parameters_to_use.append(
                dict(name="TARGET_REGION", value=self.region, type="PLAINTEXT",)
            )
            codebuild.start_build_and_wait_for_completion(
                projectName=self.project_name,
                environmentVariablesOverride=parameters_to_use,
            )
        self.write_output(self.params_for_results_display())



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class CodeBuildRunBaseTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    manifest_file_path = "manifest_file_path"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.codebuild_runs import (
            code_build_run_base_task,
        )

        self.module = code_build_run_base_task

        self.sut = self.module.CodeBuildRunBaseTask(
            manifest_file_path=self.manifest_file_path
        )

        self.wire_up_mocks()



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import luigi
from deepmerge import always_merger
import jmespath

from servicecatalog_puppet.workflow import tasks


class Boto3Task(tasks.PuppetTask):
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    client = luigi.Parameter()
    use_paginator = luigi.BoolParameter()
    call = luigi.Parameter()
    arguments = luigi.DictParameter()
    filter = luigi.Parameter()

    requester_task_id = luigi.Parameter()
    requester_task_family = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "account_id": self.account_id,
            "region": self.region,
            "client": self.client,
            "use_paginator": self.use_paginator,
            "call": self.call,
            "requester_task_id": self.requester_task_id,
            "requester_task_family": self.requester_task_family,
            "cache_invalidator": self.cache_invalidator,
        }

    def run(self):
        with self.spoke_regional_client(self.client) as client:
            if self.use_paginator:
                paginator = client.get_paginator(self.call)
                result = dict()
                for page in paginator.paginate(**self.arguments):
                    always_merger.merge(result, page)
            else:
                f = getattr(client, self.call)
                result = f(**self.arguments)

        actual_result = jmespath.search(self.filter, result)
        if isinstance(actual_result, str):
            self.write_output(actual_result.strip())
        else:
            self.write_output(actual_result)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.generic import generic_section_task
from servicecatalog_puppet.workflow.launch import launch_for_account_and_region_task
from servicecatalog_puppet.workflow.launch import launch_for_account_task
from servicecatalog_puppet.workflow.launch import launch_for_region_task
from servicecatalog_puppet.workflow.launch import launch_task


class LaunchSectionTask(generic_section_task.GenericSectionTask):
    section_name_singular = constants.LAUNCH
    section_name = constants.LAUNCHES
    for_region_task_klass = launch_for_region_task.LaunchForRegionTask
    for_account_task_klass = launch_for_account_task.LaunchForAccountTask
    for_account_and_region_task_klass = (
        launch_for_account_and_region_task.LaunchForAccountAndRegionTask
    )
    task_klass = launch_task.LaunchTask
    item_name = "launch_name"
    supports_spoke_mode = True



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet.workflow.generic import generic_for_account_and_region_task
from servicecatalog_puppet.workflow.spoke_local_portfolios import (
    spoke_local_portfolio_for_task,
)


class SpokeLocalPortfolioForAccountAndRegionTask(
    generic_for_account_and_region_task.GenericForAccountAndRegionTask,
    spoke_local_portfolio_for_task.SpokeLocalPortfolioForTask,
):
    region = luigi.Parameter()
    account_id = luigi.Parameter()



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import unittest
from copy import deepcopy
from unittest import skip

import constants


class TestManifest(unittest.TestCase):
    account_a = {
        "account_id": "012345678910",
        "default_region": "eu-west-1",
        "name": "accounta",
        "expanded_from": "ou-aaaa-aaaaaaaa",
        "organization": "o-aaaaaaaa",
        "regions_enabled": ["eu-west-2",],
        "tags": ["group:A"],
    }
    account_b = {
        "account_id": "009876543210",
        "default_region": "us-west-1",
        "expanded_from": "ou-bbbb-bbbbbbbb",
        "organization": "o-bbbbbbbb",
        "name": "accountb",
        "regions_enabled": ["us-west-2",],
        "tags": ["group:B"],
    }
    account_c = {
        "account_id": "432100098765",
        "default_region": "ap-west-1",
        "name": "accountc",
        "regions_enabled": ["ap-west-2",],
        "tags": ["group:C"],
    }
    accounts = {"accounts": [account_a, account_b, account_c,]}
    launch_a = {
        "portfolio": "portfolio_a",
        "product": "product_a",
        "execution": "hub",
        "version": "version_a",
        "deploy_to": {"tags": [{"regions": "default_region", "tag": "group:A"}]},
    }
    assertion_a = {
        "expected": {"foo": "bar"},
        "actual": {"bar": "foo"},
        "assert_for": {"tags": [{"regions": "default_region", "tag": "group:A"}]},
    }
    spoke_local_portfolio_a = {
        "portfolio": "portfolio_a",
        "deploy_to": {"tags": [{"regions": "default_region", "tag": "group:A"}]},
    }
    launch_b = {
        "portfolio": "portfolio_b",
        "product": "product_b",
        "version": "version_b",
        "execution": "hub",
        "deploy_to": {"tags": [{"regions": "default_region", "tag": "group:B"}]},
    }
    launch_c = {
        "portfolio": "portfolio_c",
        "product": "product_c",
        "version": "version_c",
        "execution": "hub",
        "deploy_to": {"tags": [{"regions": "enabled_regions", "tag": "group:B"}]},
    }
    launches = dict(
        launches=dict(launch_a=launch_a, launch_b=launch_b, launch_c=launch_c)
    )
    assertions = dict(assertions=dict(assertion_a=assertion_a))
    spoke_local_portfolios = {
        "spoke-local-portfolios": dict(spoke_local_portfolio_a=spoke_local_portfolio_a)
    }
    puppet_account_id = "0123456789010"

    def setUp(self):
        from servicecatalog_puppet.manifest_utils import Manifest

        self.sut = Manifest()

    def tearDown(self):
        self.sut = None

    def test_get_provisioning_tasks_for_launch_and_region_for_tags(self):
        # setup
        puppet_account_id = "pppppppppppp"
        section_name = "launches"
        item_name = "launch_a"
        region = "eu-west-1"
        self.sut.update(deepcopy(self.accounts))
        self.sut.update(deepcopy(self.launches))
        expected_result = [
            {
                "account_id": "012345678910",
                "account_parameters": {},
                "execution": "hub",
                "launch_name": item_name,
                "launch_parameters": {},
                "manifest_parameters": {},
                "portfolio": "portfolio_a",
                "product": "product_a",
                "puppet_account_id": puppet_account_id,
                "region": region,
                "ssm_param_outputs": [],
                "requested_priority": 0,
                "version": "version_a",
            }
        ]

        # exercise
        actual_results = self.sut.get_tasks_for_launch_and_region(
            puppet_account_id, section_name, item_name, region
        )

        # verify
        self.assertListEqual(expected_result, actual_results)

    def test_get_provisioning_tasks_for_spoke_local_portfolios_and_region_for_tags(
        self,
    ):
        # setup
        puppet_account_id = "pppppppppppp"
        section_name = "spoke-local-portfolios"
        item_name = "spoke_local_portfolio_a"
        region = "eu-west-1"
        self.sut.update(deepcopy(self.accounts))
        self.sut.update(deepcopy(self.launches))
        self.sut.update(deepcopy(self.spoke_local_portfolios))
        expected_result = [
            {
                "account_id": "012345678910",
                "spoke_local_portfolio_name": item_name,
                "portfolio": "portfolio_a",
                "puppet_account_id": puppet_account_id,
                "region": region,
                "associations": [],
                "launch_constraints": [],
                "organization": "",
                "product_generation_method": "copy",
                "sharing_mode": "ACCOUNT",
            }
        ]

        # exercise
        actual_results = self.sut.get_tasks_for_launch_and_region(
            puppet_account_id, section_name, item_name, region
        )

        # verify
        self.assertListEqual(expected_result, actual_results)

    def test_get_provisioning_tasks_for_launch_and_region_for_accounts(self):
        # setup
        puppet_account_id = "pppppppppppp"
        section_name = "launches"
        item_name = "launch_a"
        region = "eu-west-2"
        self.sut.update(deepcopy(self.accounts))
        self.sut.update(deepcopy(self.launches))
        self.sut[section_name][item_name]["deploy_to"] = dict(
            accounts=[dict(account_id="012345678910", regions="enabled_regions")]
        )
        expected_result = [
            {
                "account_id": "012345678910",
                "account_parameters": {},
                "execution": "hub",
                "launch_name": item_name,
                "launch_parameters": {},
                "manifest_parameters": {},
                "portfolio": "portfolio_a",
                "product": "product_a",
                "puppet_account_id": puppet_account_id,
                "region": region,
                "ssm_param_outputs": [],
                "requested_priority": 0,
                "version": "version_a",
            }
        ]

        # exercise
        actual_results = self.sut.get_tasks_for_launch_and_region(
            puppet_account_id, section_name, item_name, region
        )

        # verify
        self.assertListEqual(expected_result, actual_results)

    def test_get_assertion_tasks_for_launch_and_region_for_accounts(self):
        # setup
        puppet_account_id = "pppppppppppp"
        section_name = "assertions"
        item_name = "assertion_a"
        region = "eu-west-1"
        self.sut.update(deepcopy(self.accounts))
        self.sut.update(deepcopy(self.assertions))

        expected_result = [
            {
                "account_id": "012345678910",
                "assertion_name": item_name,
                "puppet_account_id": puppet_account_id,
                "region": region,
                "execution": constants.EXECUTION_MODE_DEFAULT,
                "requested_priority": 0,
                "expected": {"foo": "bar"},
                "actual": {"bar": "foo"},
            }
        ]

        # exercise
        actual_results = self.sut.get_tasks_for_launch_and_region(
            puppet_account_id, section_name, item_name, region
        )

        # verify
        self.assertListEqual(expected_result, actual_results)

    def test_get_regions_used_for_section_item_for_tags_and_default_region(self):
        # setup
        puppet_account_id = "01234567890"
        section_name = "launches"
        item_name = "launch_a"
        self.sut.update(deepcopy(self.accounts))
        self.sut.get("accounts").append(
            {
                "account_id": "9875983465794387",
                "default_region": "eu-west-3",
                "name": "accounta",
                "expanded_from": "ou-aaaa-aaaaaaaa",
                "organization": "o-aaaaaaaa",
                "regions_enabled": ["eu-west-2",],
                "tags": ["group:A"],
            }
        )
        self.sut.update(deepcopy(self.launches))
        expected_result = ["eu-west-1", "eu-west-3"]

        # exercise
        actual_results = self.sut.get_regions_used_for_section_item(
            puppet_account_id, section_name, item_name
        )
        actual_results.sort()

        # verify
        self.assertListEqual(expected_result, actual_results)

    def test_get_regions_used_for_section_item_for_accounts_and_default_region(self):
        # setup
        puppet_account_id = "01234567890"
        section_name = "launches"
        item_name = "launch_a"
        self.sut.update(deepcopy(self.accounts))
        self.sut.get("accounts").append(
            {
                "account_id": "9875983465794387",
                "default_region": "eu-west-99",
                "name": "accounta",
                "expanded_from": "ou-aaaa-aaaaaaaa",
                "organization": "o-aaaaaaaa",
                "regions_enabled": ["eu-west-2",],
                "tags": ["group:A"],
            }
        )
        self.sut.update(deepcopy(self.launches))
        self.sut["launches"][item_name]["deploy_to"] = dict(
            accounts=[dict(account_id="9875983465794387", regions="default_region")]
        )
        expected_result = ["eu-west-99"]

        # exercise
        actual_results = self.sut.get_regions_used_for_section_item(
            puppet_account_id, section_name, item_name
        )

        # verify
        self.assertListEqual(expected_result, actual_results)

    def test_get_regions_used_for_section_item_for_enabled_region(self):
        # setup
        puppet_account_id = "01234567890"
        section_name = "launches"
        item_name = "launch_c"
        self.sut.update(deepcopy(self.accounts))
        self.sut.get("accounts").append(
            {
                "account_id": "9875983465794387",
                "default_region": "eu-west-3",
                "name": "accounta",
                "expanded_from": "ou-aaaa-aaaaaaaa",
                "organization": "o-aaaaaaaa",
                "regions_enabled": ["eu-west-2",],
                "tags": ["group:A"],
            }
        )
        self.sut.update(deepcopy(self.launches))
        expected_result = ["us-west-2"]

        # exercise
        actual_results = self.sut.get_regions_used_for_section_item(
            puppet_account_id, section_name, item_name
        )

        # verify
        self.assertListEqual(expected_result, actual_results)

    def test_get_accounts_by_region(self):
        # setup
        self.sut.update(deepcopy(self.accounts))
        expected_regions = [
            "eu-west-1",
            "eu-west-2",
            "us-west-1",
            "us-west-2",
            "ap-west-1",
            "ap-west-2",
        ]

        # exercise
        actual_result = self.sut.get_accounts_by_region()
        actual_regions = list(actual_result.keys())

        # verify
        self.assertCountEqual(expected_regions, actual_regions)
        self.assertEqual(actual_result.get("eu-west-1"), [self.account_a])
        self.assertIsNone(actual_result.get("eu-west-999"))

    @skip
    def test_get_shares_by_region_portfolio_account(self):
        # setup
        self.sut.update(deepcopy(self.accounts))
        self.sut.update(deepcopy(self.launches))

        # exercise
        actual_result = self.sut.get_shares_by_region_portfolio_account(
            self.puppet_account_id, "launches"
        )

        # verify
        self.assertIsNotNone(actual_result.get("eu-west-1").get("portfolio_a"))
        self.assertEqual(
            actual_result.get("eu-west-1").get("portfolio_a").get("012345678910"),
            self.account_a,
        )
        self.assertIsNotNone(actual_result.get("us-west-1").get("portfolio_b"))
        self.assertEqual(
            actual_result.get("us-west-1").get("portfolio_b").get("009876543210"),
            self.account_b,
        )

    def test_get_sharing_policies_by_region(self):
        # setup
        self.sut.update(deepcopy(self.accounts))
        self.sut.update(deepcopy(self.launches))

        # exercise
        actual_result = self.sut.get_sharing_policies_by_region()

        # verify
        self.assertIsNotNone(actual_result.get("eu-west-1"))
        self.assertEqual(
            actual_result.get("eu-west-1"),
            dict(organizations=["o-aaaaaaaa"], accounts=[]),
        )
        self.assertEqual(
            actual_result.get("us-west-1"),
            dict(organizations=["o-bbbbbbbb"], accounts=[]),
        )
        self.assertEqual(
            actual_result.get("ap-west-1"),
            dict(organizations=[], accounts=["432100098765"]),
        )



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import json

import luigi

from servicecatalog_puppet.workflow.manifest import manifest_mixin
from servicecatalog_puppet.workflow.portfolio.accessors import (
    get_product_id_by_product_name_task,
)
from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
    portfolio_management_task,
)


class GetVersionIdByVersionName(
    portfolio_management_task.PortfolioManagementTask, manifest_mixin.ManifestMixen
):
    puppet_account_id = luigi.Parameter()
    portfolio = luigi.Parameter()
    product = luigi.Parameter()
    version = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    product_id = None
    version_id = None

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "product": self.product,
            "version": self.version,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def api_calls_used(self):
        return [
            f"servicecatalog.list_provisioning_artifacts_{self.account_id}_{self.region}",
        ]

    def requires(self):
        if self.manifest.has_cache():
            if (
                self.manifest.get("id_cache")
                .get(self.region, {})
                .get(self.portfolio, {})
                .get("products", {})
                .get(self.product, {})
            ):
                product = self.manifest["id_cache"][self.region][self.portfolio][
                    "products"
                ][self.product]
                if product.get("versions", {}).get(self.version):
                    self.product_id = product["id"]
                    self.version_id = product["versions"][self.version]
                    return []

        return dict(
            product=get_product_id_by_product_name_task.GetProductIdByProductName(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                portfolio=self.portfolio,
                product=self.product,
                account_id=self.account_id,
                region=self.region,
            )
        )

    def run(self):
        if self.product_id is not None:
            product_id = self.product_id
            version_id = self.version_id
        else:
            details = self.load_from_input("product")
            product_id = details.get("product_id")
            with self.spoke_regional_client("servicecatalog") as servicecatalog:
                version_id = None
                response = servicecatalog.list_provisioning_artifacts_single_page(
                    ProductId=product_id,
                )
                for provisioning_artifact_detail in response.get(
                    "ProvisioningArtifactDetails"
                ):
                    if provisioning_artifact_detail.get("Name") == self.version:
                        version_id = provisioning_artifact_detail.get("Id")
                assert version_id is not None, "Did not find version looking for"

        with self.output().open("w") as f:
            f.write(
                json.dumps(
                    {
                        "version_name": self.version,
                        "version_id": version_id,
                        "product_name": self.product,
                        "product_id": product_id,
                    },
                    indent=4,
                    default=str,
                )
            )



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow import tasks
from servicecatalog_puppet.workflow.manifest import manifest_mixin


class SpokeLocalPortfolioBaseTask(tasks.PuppetTask, manifest_mixin.ManifestMixen):
    manifest_file_path = luigi.Parameter()

    @property
    def status(self):
        return (
            self.manifest.get(self.section_name)
            .get(self.spoke_local_portfolio_name)
            .get("status", constants.SPOKE_LOCAL_PORTFOLIO_STATUS_SHARED)
        )

    @property
    def section_name(self):
        return constants.SPOKE_LOCAL_PORTFOLIOS

    @property
    def item_name(self):
        return self.spoke_local_portfolio_name

    @property
    def item_identifier(self):
        return "spoke_local_portfolio_name"



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.launch import provision_product_dry_run_task
from servicecatalog_puppet.workflow.launch import provision_product_task
from servicecatalog_puppet.workflow.launch import provisioning_task
from servicecatalog_puppet.workflow.launch import terminate_product_dry_run_task
from servicecatalog_puppet.workflow.launch import terminate_product_task


class LaunchForTask(provisioning_task.ProvisioningTask):
    launch_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "launch_name": self.launch_name,
            "cache_invalidator": self.cache_invalidator,
        }

    def get_klass_for_provisioning(self):
        if self.is_dry_run:
            if self.status == constants.PROVISIONED:
                return provision_product_dry_run_task.ProvisionProductDryRunTask
            elif self.status == constants.TERMINATED:
                return terminate_product_dry_run_task.TerminateProductDryRunTask
            else:
                raise Exception(f"Unknown status: {self.status}")
        else:
            if self.status == constants.PROVISIONED:
                return provision_product_task.ProvisionProductTask
            elif self.status == constants.TERMINATED:
                return terminate_product_task.TerminateProductTask
            else:
                raise Exception(f"Unknown status: {self.status}")

    def run(self):
        self.write_output(self.params_for_results_display())



def get_accounts_for_path(client, path):
    ou = client.convert_path_to_ou(path)
    response = client.list_children_nested(ParentId=ou, ChildType="ACCOUNT")
    return ",".join([r.get("Id") for r in response])


#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

macros = {"get_accounts_for_path": get_accounts_for_path}



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import troposphere as t
from troposphere import iam

from servicecatalog_puppet import config
from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow import tasks


class CreateTemplateForStackTask(tasks.PuppetTask):
    def params_for_results_display(self):
        return {
            "cache_invalidator": self.cache_invalidator,
        }

    def run(self):
        puppet_version = constants.VERSION
        description = f"""Bootstrap template used to configure spoke account for stack use
        {{"version": "{puppet_version}", "framework": "servicecatalog-puppet", "role": "bootstrap-spoke-stack"}}"""

        template = t.Template(Description=description)

        template.add_resource(
            iam.Role(
                "PuppetStackRole",
                RoleName="PuppetStackRole",
                ManagedPolicyArns=[
                    t.Sub("arn:${AWS::Partition}:iam::aws:policy/AdministratorAccess")
                ],
                Path=config.get_puppet_role_path(),
                AssumeRolePolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": ["sts:AssumeRole"],
                            "Effect": "Allow",
                            "Principal": {"Service": ["cloudformation.amazonaws.com"]},
                        }
                    ],
                },
            )
        )

        self.write_output(template.to_yaml(), skip_json_dump=True)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.generic import generic_section_task
from servicecatalog_puppet.workflow.workspaces import (
    workspace_for_account_and_region_task,
)
from servicecatalog_puppet.workflow.workspaces import workspace_for_account_task
from servicecatalog_puppet.workflow.workspaces import workspace_for_region_task
from servicecatalog_puppet.workflow.workspaces import workspace_task


class WorkspaceSectionTask(generic_section_task.GenericSectionTask):
    """
    Iterates through the workspace items in the workspace section of the manifest file and generates the dependencies
    tasks and correct ForX tasks.
    """

    section_name_singular = constants.WORKSPACE
    section_name = constants.WORKSPACES
    for_region_task_klass = workspace_for_region_task.WorkspaceForRegionTask
    for_account_task_klass = workspace_for_account_task.WorkspaceForAccountTask
    for_account_and_region_task_klass = (
        workspace_for_account_and_region_task.WorkspaceForAccountAndRegionTask
    )
    task_klass = workspace_task.WorkspaceTask
    item_name = "workspace_name"
    supports_spoke_mode = True



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import config
from servicecatalog_puppet.workflow.launch import provisioning_task


class ResetProvisionedProductOwnerTask(provisioning_task.ProvisioningTask):
    launch_name = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "launch_name": self.launch_name,
            "account_id": self.account_id,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
        }

    def api_calls_used(self):
        return [
            f"servicecatalog.scan_provisioned_products_single_page_{self.account_id}_{self.region}",
            f"servicecatalog.update_provisioned_product_properties_{self.account_id}_{self.region}",
        ]

    def run(self):
        self.info(f"starting ResetProvisionedProductOwnerTask")

        with self.spoke_regional_client("servicecatalog") as service_catalog:
            self.info(f"Checking if existing provisioned product exists")
            all_results = service_catalog.scan_provisioned_products_single_page(
                AccessLevelFilter={"Key": "Account", "Value": "self"},
            ).get("ProvisionedProducts", [])
            changes_made = list()
            for result in all_results:
                if result.get("Name") == self.launch_name:
                    provisioned_product_id = result.get("Id")
                    self.info(f"Ensuring current provisioned product owner is correct")
                    changes_made.append(result)
                    service_catalog.update_provisioned_product_properties(
                        ProvisionedProductId=provisioned_product_id,
                        ProvisionedProductProperties={
                            "OWNER": config.get_puppet_role_arn(self.account_id)
                        },
                    )
            self.write_output(changes_made)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.generic import generic_section_task
from servicecatalog_puppet.workflow.spoke_local_portfolios import (
    spoke_local_portfolio_for_account_and_region_task,
)
from servicecatalog_puppet.workflow.spoke_local_portfolios import (
    spoke_local_portfolio_for_account_task,
)
from servicecatalog_puppet.workflow.spoke_local_portfolios import (
    spoke_local_portfolio_for_region_task,
)
from servicecatalog_puppet.workflow.spoke_local_portfolios import (
    spoke_local_portfolio_task,
)


class SpokeLocalPortfolioSectionTask(generic_section_task.GenericSectionTask):
    section_name_singular = constants.SPOKE_LOCAL_PORTFOLIO
    section_name = constants.SPOKE_LOCAL_PORTFOLIOS
    for_region_task_klass = (
        spoke_local_portfolio_for_region_task.SpokeLocalPortfolioForRegionTask
    )
    for_account_task_klass = (
        spoke_local_portfolio_for_account_task.SpokeLocalPortfolioForAccountTask
    )
    for_account_and_region_task_klass = (
        spoke_local_portfolio_for_account_and_region_task.SpokeLocalPortfolioForAccountAndRegionTask
    )
    task_klass = spoke_local_portfolio_task.SpokeLocalPortfolioTask
    item_name = "spoke_local_portfolio_name"
    supports_spoke_mode = False



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import json
import logging
import math
import os
import traceback
from pathlib import Path

import luigi
import psutil
from betterboto import client as betterboto_client
from luigi import format
from luigi.contrib import s3

from servicecatalog_puppet import constants, config

logger = logging.getLogger("tasks")
logger.setLevel(logging.INFO)


def unwrap(what):
    if hasattr(what, "get_wrapped"):
        thing = what.get_wrapped()
    else:
        thing = what
    if isinstance(what, dict):
        for k, v in thing.items():
            thing[k] = unwrap(v)
    return thing


class PuppetTask(luigi.Task):
    @property
    def executor_account_id(self):
        return os.environ.get("EXECUTOR_ACCOUNT_ID")

    @property
    def execution_mode(self):
        return os.environ.get("SCT_EXECUTION_MODE", constants.EXECUTION_MODE_HUB)

    def is_running_in_spoke(self):
        return self.execution_mode == constants.EXECUTION_MODE_SPOKE

    @property
    def single_account(self):
        return os.environ.get("SCT_SINGLE_ACCOUNT", "None")

    @property
    def should_use_product_plans(self):
        if self.execution_mode == constants.EXECUTION_MODE_HUB:
            return os.environ.get("SCT_SHOULD_USE_PRODUCT_PLANS", "True") == "True"
        else:
            return False

    @property
    def cache_invalidator(self):
        return os.environ.get("SCT_CACHE_INVALIDATOR", "NOW")

    @property
    def is_dry_run(self):
        return os.environ.get("SCT_IS_DRY_RUN", "False") == "True"

    @property
    def should_use_sns(self):
        return os.environ.get("SCT_SHOULD_USE_SNS", "False") == "True"

    def get_account_used(self):
        return self.account_id if self.is_running_in_spoke() else self.puppet_account_id

    def spoke_client(self, service):
        return betterboto_client.CrossAccountClientContextManager(
            service,
            config.get_puppet_role_arn(self.account_id),
            f"{self.account_id}-{config.get_puppet_role_name()}",
        )

    def spoke_regional_client(self, service, region_name=None):
        region = region_name or self.region
        return betterboto_client.CrossAccountClientContextManager(
            service,
            config.get_puppet_role_arn(self.account_id),
            f"{self.account_id}-{self.region}-{config.get_puppet_role_name()}",
            region_name=region,
        )

    def hub_client(self, service):
        if self.is_running_in_spoke():
            return betterboto_client.CrossAccountClientContextManager(
                service,
                config.get_puppet_role_arn(self.executor_account_id),
                f"{self.executor_account_id}-{config.get_puppet_role_name()}",
            )
        else:
            return betterboto_client.CrossAccountClientContextManager(
                service,
                config.get_puppet_role_arn(self.puppet_account_id),
                f"{self.puppet_account_id}-{config.get_puppet_role_name()}",
            )

    def hub_regional_client(self, service, region_name=None):
        region = region_name or self.region
        if self.is_running_in_spoke():
            return betterboto_client.CrossAccountClientContextManager(
                service,
                config.get_puppet_role_arn(self.executor_account_id),
                f"{self.executor_account_id}-{config.get_puppet_role_name()}",
                region_name=region,
            )
        else:
            return betterboto_client.CrossAccountClientContextManager(
                service,
                config.get_puppet_role_arn(self.puppet_account_id),
                f"{self.puppet_account_id}-{region}-{config.get_puppet_role_name()}",
                region_name=region,
            )

    def read_from_input(self, input_name):
        with self.input().get(input_name).open("r") as f:
            return f.read()

    def load_from_input(self, input_name):
        return json.loads(self.read_from_input(input_name))

    def info(self, message):
        logger.info(f"{self.uid}: {message}")

    def error(self, message):
        logger.error(f"{self.uid}: {message}")

    def warning(self, message):
        logger.warning(f"{self.uid}: {message}")

    def api_calls_used(self):
        return []

    def resources_used(self):
        return []

    @property
    def resources(self):
        result = {}
        api_calls = self.api_calls_used()
        if isinstance(api_calls, list):
            for a in self.api_calls_used():
                result[a] = 1
        elif isinstance(api_calls, dict):
            for a, r in api_calls.items():
                result[a] = r

        for i, r in self.resources_used():
            result[f"{i}-{r.name}"] = 1 / r.value
        return result

    @property
    def output_location(self):
        puppet_account_id = config.get_puppet_account_id()
        path = f"output/{self.uid}.{self.output_suffix}"
        should_use_s3_target_if_caching_is_on = (
            "cache_invalidator" not in self.params_for_results_display().keys()
        )
        if should_use_s3_target_if_caching_is_on and config.is_caching_enabled(
            puppet_account_id
        ):
            return f"s3://sc-puppet-caching-bucket-{config.get_puppet_account_id()}-{config.get_home_region(puppet_account_id)}/{path}"
        else:
            return path

    def output(self):
        should_use_s3_target_if_caching_is_on = (
            "cache_invalidator" not in self.params_for_results_display().keys()
        )
        if should_use_s3_target_if_caching_is_on and config.is_caching_enabled(
            config.get_puppet_account_id()
        ):
            return s3.S3Target(self.output_location, format=format.UTF8)
        else:
            return luigi.LocalTarget(self.output_location)

    @property
    def output_suffix(self):
        return "json"

    @property
    def uid(self):
        return f"{self.__class__.__name__}/{self.node_id}"

    def params_for_results_display(self):
        return {}

    def write_output(self, content, skip_json_dump=False):
        with self.output().open("w") as f:
            if skip_json_dump:
                f.write(content)
            else:
                f.write(json.dumps(content, indent=4, default=str,))

    @property
    def node_id(self):
        values = [self.__class__.__name__.replace("Task", "")] + [
            str(v) for v in self.params_for_results_display().values()
        ]
        return "/".join(values)

    def graph_node(self):
        task_friendly_name = self.__class__.__name__.replace("Task", "")
        task_description = ""
        for param, value in self.params_for_results_display().items():
            task_description += f"<br/>{param}: {value}"
        label = f"<b>{task_friendly_name}</b>{task_description}"
        return f'"{self.node_id}" [fillcolor=lawngreen style=filled label= < {label} >]'


def record_event(event_type, task, extra_event_data=None):
    task_type = task.__class__.__name__
    task_params = task.param_kwargs

    event = {
        "event_type": event_type,
        "task_type": task_type,
        "task_params": task_params,
        "params_for_results": task.params_for_results_display(),
    }
    if extra_event_data is not None:
        event.update(extra_event_data)

    with open(
        Path(constants.RESULTS_DIRECTORY)
        / event_type
        / f"{task_type}-{task.task_id}.json",
        "w",
    ) as f:
        f.write(json.dumps(event, default=str, indent=4,))


@luigi.Task.event_handler(luigi.Event.FAILURE)
def on_task_failure(task, exception):
    exception_details = {
        "exception_type": type(exception),
        "exception_stack_trace": traceback.format_exception(
            etype=type(exception), value=exception, tb=exception.__traceback__,
        ),
    }
    record_event("failure", task, exception_details)


def print_stats():
    mem = psutil.virtual_memory()
    logger.info(
        f"memory usage: total={math.ceil(mem.total / 1024 / 1024)}MB used={math.ceil(mem.used / 1024 / 1024)}MB percent={mem.percent}%"
    )


@luigi.Task.event_handler(luigi.Event.START)
def on_task_success(task):
    task_name = task.__class__.__name__
    to_string = f"{task_name}: "
    for name, value in task.param_kwargs.items():
        to_string += f"{name}={value}, "
    logger.info(f"{to_string} started")
    record_event("start", task)


@luigi.Task.event_handler(luigi.Event.SUCCESS)
def on_task_success(task):
    print_stats()
    record_event("success", task)


@luigi.Task.event_handler(luigi.Event.TIMEOUT)
def on_task_timeout(task):
    print_stats()
    record_event("timeout", task)


@luigi.Task.event_handler(luigi.Event.PROCESS_FAILURE)
def on_task_process_failure(task, error_msg):
    print_stats()
    exception_details = {
        "exception_type": "PROCESS_FAILURE",
        "exception_stack_trace": error_msg,
    }
    record_event("process_failure", task, exception_details)


@luigi.Task.event_handler(luigi.Event.PROCESSING_TIME)
def on_task_processing_time(task, duration):
    print_stats()
    record_event("processing_time", task, {"duration": duration})

    task_params = dict(**task.param_kwargs)
    task_params.update(task.params_for_results_display())

    with betterboto_client.CrossAccountClientContextManager(
        "cloudwatch",
        config.get_puppet_role_arn(config.get_puppet_account_id()),
        "cloudwatch-puppethub",
    ) as cloudwatch:

        dimensions = [dict(Name="task_type", Value=task.__class__.__name__,)]
        for note_worthy in [
            "launch_name",
            "region",
            "account_id",
            "puppet_account_id",
            "sharing_mode",
            "portfolio",
            "product",
            "version",
            "execution",
        ]:
            if task_params.get(note_worthy):
                dimensions.append(
                    dict(Name=str(note_worthy), Value=str(task_params.get(note_worthy)))
                )

        cloudwatch.put_metric_data(
            Namespace=f"ServiceCatalogTools/Puppet/v1/ProcessingTime/{task.__class__.__name__}",
            MetricData=[
                dict(
                    MetricName=task.__class__.__name__,
                    Dimensions=dimensions,
                    Value=duration,
                    Unit="Seconds",
                ),
            ],
        )
        cloudwatch.put_metric_data(
            Namespace=f"ServiceCatalogTools/Puppet/v1/ProcessingTime/Tasks",
            MetricData=[
                dict(
                    MetricName="Tasks",
                    Dimensions=[dict(Name="TaskType", Value=task.__class__.__name__)],
                    Value=duration,
                    Unit="Seconds",
                ),
            ],
        )


@luigi.Task.event_handler(luigi.Event.BROKEN_TASK)
def on_task_broken_task(task, exception):
    print_stats()
    record_event("broken_task", task, {"exception": exception})



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0


def test_constants_values():
    # setup
    from servicecatalog_puppet import constants

    # exercise
    # verify
    assert constants.PREFIX == "SC-P--"
    assert constants.BOOTSTRAP_STACK_NAME == "servicecatalog-puppet"
    assert constants.PIPELINE_NAME == "servicecatalog-puppet-pipeline"
    assert constants.SERVICE_CATALOG_PUPPET_REPO_NAME == "ServiceCatalogPuppet"
    assert constants.OUTPUT == "output"

    assert constants.CONFIG_PARAM_NAME == "/servicecatalog-puppet/config"
    assert (
        constants.CONFIG_PARAM_NAME_ORG_IAM_ROLE_ARN
        == "/servicecatalog-puppet/org-iam-role-arn"
    )
    assert constants.PUPPET_ORG_ROLE_FOR_EXPANDS_ARN == "PuppetOrgRoleForExpandsArn"
    assert constants.HOME_REGION_PARAM_NAME == "/servicecatalog-puppet/home-region"

    assert constants.PROVISIONED == "provisioned"
    assert constants.TERMINATED == "terminated"

    assert constants.DEFAULT_TIMEOUT == 0
    assert constants.PARAMETERS == "parameters"
    assert constants.ACCOUNTS == "accounts"
    assert constants.MAPPINGS == "mappings"
    assert constants.LAUNCHES == "launches"
    assert constants.STACKS == "stacks"
    assert constants.SPOKE_LOCAL_PORTFOLIOS == "spoke-local-portfolios"
    assert constants.LAMBDA_INVOCATIONS == "lambda-invocations"
    assert constants.ASSERTIONS == "assertions"
    assert constants.ASSERTION == "assertion"
    assert constants.CODE_BUILD_RUNS == "code-build-runs"
    assert constants.CODE_BUILD_RUN == "code-build-run"
    assert constants.ACTIONS == "actions"

    assert constants.LAUNCH == "launch"
    assert constants.STACK == "stack"
    assert constants.LAMBDA_INVOCATION == "lambda-invocation"
    assert constants.SPOKE_LOCAL_PORTFOLIO == "spoke-local-portfolio"
    assert constants.ACTION == "action"

    assert constants.AFFINITY == "affinity"
    assert constants.AFFINITY_REGION == "region"
    assert constants.AFFINITY_ACCOUNT == "account"
    assert constants.AFFINITY_ACCOUNT_AND_REGION == "account-and-region"

    assert constants.ALL_SECTION_NAMES == [
        constants.LAUNCHES,
        constants.STACKS,
        constants.SPOKE_LOCAL_PORTFOLIOS,
        constants.LAMBDA_INVOCATIONS,
        constants.CODE_BUILD_RUNS,
        constants.ASSERTIONS,
        constants.APPS,
        constants.WORKSPACES,
    ]
    assert constants.RESULTS_DIRECTORY == "results"
    assert constants.NO_CHANGE == "NO_CHANGE"
    assert constants.CHANGE == "CHANGE"
    assert constants.EVENT_BUS_NAME == "servicecatalog-puppet-event-bus"
    assert constants.EVENT_BUS_IN_SPOKE_NAME == "servicecatalog-puppet-spoke-event-bus"
    assert constants.SERVICE_CATALOG_PUPPET_EVENT_SOURCE == "servicecatalog-puppet"
    assert constants.SERVICE_CATALOG_PUPPET_OPS_CENTER_SOURCE == "servicecatalog-puppet"

    assert constants.EVENTBRIDGE_MAX_EVENTS_PER_CALL == 10
    assert (
        constants.SPOKE_VERSION_SSM_PARAM_NAME == "service-catalog-puppet-spoke-version"
    )
    assert constants.PUPPET_VERSION_SSM_PARAM_NAME == "service-catalog-puppet-version"
    assert (
        constants.PUPPET_VERSION_INITIAL_SSM_PARAM_NAME
        == "service-catalog-puppet-version-initial"
    )
    assert constants.SPOKE_LOCAL_PORTFOLIO_STATUS_SHARED == "shared"
    assert constants.SPOKE_LOCAL_PORTFOLIO_STATUS_TERMINATED == "terminated"
    assert constants.EXECUTION_MODE_ASYNC == "async"
    assert constants.EXECUTION_MODE_HUB == "hub"
    assert constants.EXECUTION_MODE_SPOKE == "spoke"
    assert constants.EXECUTION_MODE_DEFAULT == constants.EXECUTION_MODE_HUB
    assert (
        constants.EXECUTION_SPOKE_CODEBUILD_PROJECT_NAME
        == "servicecatalog-puppet-deploy-in-spoke"
    )
    assert constants.SHARING_MODE_ACCOUNT == "ACCOUNT"
    assert constants.SHARING_MODE_AWS_ORGANIZATIONS == "AWS_ORGANIZATIONS"
    assert constants.SHARING_MODE_DEFAULT == constants.SHARING_MODE_ACCOUNT
    assert constants.PARTITION_DEFAULT == "aws"
    assert constants.PARTITION_ENVIRONMENTAL_VARIABLE_NAME == "PARTITION"
    assert (
        constants.PARTITION_SSM_PARAMETER_VARIABLE_NAME
        == "/servicecatalog-puppet/partition"
    )
    assert constants.PUPPET_ROLE_NAME_DEFAULT == "PuppetRole"
    assert constants.PUPPET_ROLE_NAME_ENVIRONMENTAL_VARIABLE_NAME == "PUPPET_ROLE_NAME"
    assert (
        constants.PUPPET_ROLE_NAME_SSM_PARAMETER_VARIABLE_NAME
        == "/servicecatalog-puppet/puppet-role/name"
    )
    assert constants.PUPPET_ROLE_PATH_DEFAULT == "/servicecatalog-puppet/"
    assert constants.PUPPET_ROLE_PATH_ENVIRONMENTAL_VARIABLE_NAME == "PUPPET_ROLE_PATH"
    assert (
        constants.PUPPET_ROLE_PATH_SSM_PARAMETER_VARIABLE_NAME
        == "/servicecatalog-puppet/puppet-role/path"
    )
    assert constants.DEPLOY_ENVIRONMENT_COMPUTE_TYPE_DEFAULT == "BUILD_GENERAL1_SMALL"
    assert constants.CONFIG_IS_CACHING_ENABLED == "CONFIG_IS_CACHING_ENABLED"
    assert (
        constants.START_SHARED_SCHEDULER_COMMAND
        == "luigid --background --pidfile luigi.pid --logdir results/logs --state-path results/state"
    )
    assert constants.CONFIG_REGIONS == "regions"
    assert (
        constants.CONFIG_SHOULD_COLLECT_CLOUDFORMATION_EVENTS
        == "should_collect_cloudformation_events"
    )
    assert constants.CONFIG_SHOULD_USE_SHARED_SCHEDULER == "should_use_shared_scheduler"
    assert constants.CONFIG_SHOULD_EXPLODE_MANIFEST == "should_explode_manifest"

    assert constants.VERSION_OVERRIDE == "SCP_VERSION_OVERRIDE"

    assert constants.PRODUCT_GENERATION_METHOD_DEFAULT == "copy"
    assert constants.LUIGI_DEFAULT_LOG_LEVEL == "INFO"
    assert constants.ALL_SECTION_NAME_SINGULAR_AND_PLURAL_LIST == [
        (constants.LAUNCH, constants.LAUNCHES),
        (constants.STACK, constants.STACKS),
        (constants.SPOKE_LOCAL_PORTFOLIO, constants.SPOKE_LOCAL_PORTFOLIOS),
        (constants.LAMBDA_INVOCATION, constants.LAMBDA_INVOCATIONS),
        (constants.CODE_BUILD_RUN, constants.CODE_BUILD_RUNS),
        (constants.ASSERTION, constants.ASSERTIONS),
        (constants.APP, constants.APPS),
        (constants.WORKSPACE, constants.WORKSPACES),
    ]
    assert constants.SECTION_NAME_SINGULAR_AND_PLURAL_LIST_THAT_SUPPORTS_PARAMETERS == [
        (constants.LAUNCH, constants.LAUNCHES),
        (constants.STACK, constants.STACKS),
        (constants.LAMBDA_INVOCATION, constants.LAMBDA_INVOCATIONS),
        (constants.CODE_BUILD_RUN, constants.CODE_BUILD_RUNS),
        (constants.APP, constants.APPS),
        (constants.WORKSPACE, constants.WORKSPACES),
    ]
    assert constants.SECTION_SINGULAR_TO_PLURAL == {
        constants.LAUNCH: constants.LAUNCHES,
        constants.STACK: constants.STACKS,
        constants.SPOKE_LOCAL_PORTFOLIO: constants.SPOKE_LOCAL_PORTFOLIOS,
        constants.LAMBDA_INVOCATION: constants.LAMBDA_INVOCATIONS,
        constants.CODE_BUILD_RUN: constants.CODE_BUILD_RUNS,
        constants.ASSERTION: constants.ASSERTIONS,
        constants.APP: constants.APPS,
        constants.WORKSPACE: constants.WORKSPACES,
    }



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0


def slugify_for_cloudformation_stack_name(raw) -> str:
    return raw.replace("_", "-")



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import copy
import json

import luigi
import yaml

from servicecatalog_puppet import config
from servicecatalog_puppet.workflow import tasks
from servicecatalog_puppet.workflow.manifest import manifest_mixin
from servicecatalog_puppet.workflow.portfolio.accessors import (
    get_portfolio_by_portfolio_name_task,
)
from servicecatalog_puppet.workflow.portfolio.accessors import (
    get_products_and_provisioning_artifacts_task,
)


class GenerateManifestWithIdsTask(tasks.PuppetTask, manifest_mixin.ManifestMixen):
    puppet_account_id = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        requirements = dict()
        regions = config.get_regions(self.puppet_account_id)
        for launch_name, launch_details in self.manifest.get_launches_items():
            portfolio = launch_details.get("portfolio")
            for region in regions:
                if requirements.get(region) is None:
                    requirements[region] = dict()

                regional_details = requirements[region]
                if regional_details.get(portfolio) is None:
                    regional_details[portfolio] = dict(products=dict())

                portfolio_details = regional_details[portfolio]
                if portfolio_details.get("details") is None:
                    portfolio_details[
                        "details"
                    ] = get_portfolio_by_portfolio_name_task.GetPortfolioByPortfolioName(
                        manifest_file_path=self.manifest_file_path,
                        portfolio=portfolio,
                        puppet_account_id=self.puppet_account_id,
                        account_id=self.puppet_account_id,
                        region=region,
                    )

                product = launch_details.get("product")
                products = portfolio_details.get("products")
                if products.get(product) is None:
                    products[
                        product
                    ] = get_products_and_provisioning_artifacts_task.GetProductsAndProvisioningArtifactsTask(
                        manifest_file_path=self.manifest_file_path,
                        region=region,
                        portfolio=portfolio,
                        puppet_account_id=self.puppet_account_id,
                    )
        return requirements

    def run(self):
        new_manifest = copy.deepcopy(self.manifest)
        regions = config.get_regions(self.puppet_account_id)
        global_id_cache = dict()
        new_manifest["id_cache"] = global_id_cache

        for region in regions:
            regional_id_cache = dict()
            r = self.input().get(region)
            for launch_name, launch_details in self.manifest.get_launches_items():
                target = r.get(launch_details.get("portfolio")).get("details")
                portfolio_id = json.loads(target.open("r").read()).get("portfolio_id")
                portfolio_name = launch_details.get("portfolio")
                if regional_id_cache.get(portfolio_name) is None:
                    regional_id_cache[portfolio_name] = dict(
                        id=portfolio_id, products=dict()
                    )

                product = launch_details.get("product")
                target = (
                    r.get(launch_details.get("portfolio")).get("products").get(product)
                )
                all_details = json.loads(target.open("r").read())
                all_products_and_their_versions = all_details
                for p in all_products_and_their_versions:
                    if (
                        regional_id_cache[portfolio_name]["products"].get(p.get("Name"))
                        is None
                    ):
                        regional_id_cache[portfolio_name]["products"][
                            p.get("Name")
                        ] = dict(id=p.get("ProductId"), versions=dict())

                    for a in p.get("provisioning_artifact_details"):
                        regional_id_cache[portfolio_name]["products"][p.get("Name")][
                            "versions"
                        ][a.get("Name")] = a.get("Id")

            global_id_cache[region] = regional_id_cache

        self.write_output(
            yaml.safe_dump(json.loads(json.dumps(new_manifest))), skip_json_dump=True
        )



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow import dependency
from servicecatalog_puppet.workflow.general import get_ssm_param_task
from servicecatalog_puppet.workflow.stack import provisioning_task


class TerminateStackTask(
    provisioning_task.ProvisioningTask,
    dependency.DependenciesMixin,
    get_ssm_param_task.PuppetTaskWithParameters,
):
    stack_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    bucket = luigi.Parameter()
    key = luigi.Parameter()
    version_id = luigi.Parameter()

    capabilities = luigi.ListParameter()

    use_service_role = luigi.BoolParameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    retry_count = luigi.IntParameter(default=1, significant=False)
    worker_timeout = luigi.IntParameter(default=0, significant=False)
    ssm_param_outputs = luigi.ListParameter(default=[], significant=False)
    requested_priority = luigi.IntParameter(significant=False, default=0)

    execution = luigi.Parameter()

    try_count = 1

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "stack_name": self.stack_name,
            "account_id": self.account_id,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        requirements = {"section_dependencies": self.get_section_dependencies()}
        return requirements

    def run(self):
        if self.execution == constants.EXECUTION_MODE_HUB:
            self.terminate_ssm_outputs()

        with self.spoke_regional_client("cloudformation") as cloudformation:
            cloudformation.ensure_deleted(StackName=self.stack_name)

        self.write_output(self.params_for_results_display())



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import click
import pkg_resources
from betterboto import client as betterboto_client


def version():
    click.echo(
        "cli version: {}".format(
            pkg_resources.require("aws-service-catalog-puppet")[0].version
        )
    )
    with betterboto_client.ClientContextManager("ssm") as ssm:
        response = ssm.get_parameter(Name="service-catalog-puppet-regional-version")
        click.echo(
            "regional stack version: {} for region: {}".format(
                response.get("Parameter").get("Value"),
                response.get("Parameter").get("ARN").split(":")[3],
            )
        )
        response = ssm.get_parameter(Name="service-catalog-puppet-version")
        click.echo("stack version: {}".format(response.get("Parameter").get("Value"),))



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet.workflow.portfolio.accessors import (
    search_products_as_admin_task,
)
from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
    portfolio_management_task,
)


class GetProductsAndProvisioningArtifactsTask(
    portfolio_management_task.PortfolioManagementTask
):
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        return {
            "search_products_as_admin": search_products_as_admin_task.SearchProductsAsAdminTask(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                portfolio=self.portfolio,
                region=self.region,
                account_id=self.puppet_account_id,
            )
        }

    def api_calls_used(self):
        return [
            f"servicecatalog.list_provisioning_artifacts_{self.puppet_account_id}_{self.region}",
        ]

    def run(self):
        product_and_artifact_details = []
        with self.hub_regional_client("servicecatalog") as service_catalog:
            response = self.load_from_input("search_products_as_admin")
            for product_view_detail in response.get("ProductViewDetails", []):
                product_view_summary = product_view_detail.get("ProductViewSummary")
                product_view_summary["ProductARN"] = product_view_detail.get(
                    "ProductARN"
                )
                product_and_artifact_details.append(product_view_summary)

                provisioning_artifact_details = product_view_summary[
                    "provisioning_artifact_details"
                ] = []
                hub_product_id = product_view_summary.get("ProductId")
                hub_provisioning_artifact_details = service_catalog.list_provisioning_artifacts(
                    ProductId=hub_product_id
                ).get(
                    "ProvisioningArtifactDetails", []
                )
                for (
                    hub_provisioning_artifact_detail
                ) in hub_provisioning_artifact_details:
                    if (
                        hub_provisioning_artifact_detail.get("Type")
                        == "CLOUD_FORMATION_TEMPLATE"
                    ):
                        provisioning_artifact_details.append(
                            hub_provisioning_artifact_detail
                        )

        self.write_output(product_and_artifact_details)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import deepdiff
from datetime import datetime
import luigi


from servicecatalog_puppet.workflow import dependency
from servicecatalog_puppet.workflow.assertions import assertion_base_task
from servicecatalog_puppet.workflow.manifest import manifest_mixin
from servicecatalog_puppet.workflow.general import boto3_task


class DoAssertTask(
    assertion_base_task.AssertionBaseTask,
    manifest_mixin.ManifestMixen,
    dependency.DependenciesMixin,
):
    assertion_name = luigi.Parameter()
    region = luigi.Parameter()
    account_id = luigi.Parameter()

    puppet_account_id = luigi.Parameter()
    execution = luigi.Parameter()

    expected = luigi.DictParameter()
    actual = luigi.DictParameter()

    requested_priority = luigi.IntParameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "assertion_name": self.assertion_name,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        config = self.actual.get("config")
        return dict(
            result=boto3_task.Boto3Task(
                account_id=self.account_id,
                region=self.region,
                client=config.get("client"),
                use_paginator=config.get("use_paginator"),
                call=config.get("call"),
                arguments=config.get("arguments"),
                filter=config.get("filter"),
                requester_task_id=self.task_id,
                requester_task_family=self.task_family,
            )
        )

    def run(self):
        actual_result = self.load_from_input("result")
        expected_result = self.expected.get("config").get("value")
        if isinstance(expected_result, tuple):
            expected_result = list(expected_result)
        elif isinstance(actual_result, str):
            expected_result = expected_result.strip()

        ddiff = deepdiff.DeepDiff(actual_result, expected_result, ignore_order=True)
        if len(ddiff.keys()) > 0:
            raise Exception(ddiff)
        else:
            self.write_output(self.params_for_results_display())



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.generic import generic_section_task
from servicecatalog_puppet.workflow.stack import stack_for_account_and_region_task
from servicecatalog_puppet.workflow.stack import stack_for_account_task
from servicecatalog_puppet.workflow.stack import stack_for_region_task
from servicecatalog_puppet.workflow.stack import stack_task


class StackSectionTask(generic_section_task.GenericSectionTask):
    section_name_singular = constants.STACK
    section_name = constants.STACKS
    for_region_task_klass = stack_for_region_task.StackForRegionTask
    for_account_task_klass = stack_for_account_task.StackForAccountTask
    for_account_and_region_task_klass = (
        stack_for_account_and_region_task.StackForAccountAndRegionTask
    )
    task_klass = stack_task.StackTask
    item_name = "stack_name"
    supports_spoke_mode = True



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.generic import generic_section_task
from servicecatalog_puppet.workflow.lambda_invocations import (
    lambda_invocation_for_account_and_region_task,
)
from servicecatalog_puppet.workflow.lambda_invocations import (
    lambda_invocation_for_account_task,
)
from servicecatalog_puppet.workflow.lambda_invocations import (
    lambda_invocation_for_region_task,
)
from servicecatalog_puppet.workflow.lambda_invocations import lambda_invocation_task


class LambdaInvocationsSectionTask(generic_section_task.GenericSectionTask):
    section_name_singular = constants.LAMBDA_INVOCATION
    section_name = constants.LAMBDA_INVOCATIONS
    for_region_task_klass = (
        lambda_invocation_for_region_task.LambdaInvocationForRegionTask
    )
    for_account_task_klass = (
        lambda_invocation_for_account_task.LambdaInvocationForAccountTask
    )
    for_account_and_region_task_klass = (
        lambda_invocation_for_account_and_region_task.LambdaInvocationForAccountAndRegionTask
    )
    task_klass = lambda_invocation_task.LambdaInvocationTask
    item_name = "lambda_invocation_name"
    supports_spoke_mode = True



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet.workflow.portfolio.associations import (
    create_associations_in_python_for_portfolio_task,
)
from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
    portfolio_management_task,
)
from servicecatalog_puppet.workflow.portfolio.sharing_management import (
    share_and_accept_portfolio_task,
)


class CreateShareForAccountLaunchRegion(
    portfolio_management_task.PortfolioManagementTask
):
    """for the given account_id and launch and region create the shares"""

    puppet_account_id = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    sharing_mode = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "sharing_mode": self.sharing_mode,
            "account_id": self.account_id,
        }

    def requires(self):
        if self.account_id == self.puppet_account_id:
            return create_associations_in_python_for_portfolio_task.CreateAssociationsInPythonForPortfolioTask(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                account_id=self.account_id,
                region=self.region,
                portfolio=self.portfolio,
            )
        else:
            return share_and_accept_portfolio_task.ShareAndAcceptPortfolioTask(
                manifest_file_path=self.manifest_file_path,
                account_id=self.account_id,
                region=self.region,
                portfolio=self.portfolio,
                puppet_account_id=self.puppet_account_id,
                sharing_mode=self.sharing_mode,
            )

    def run(self):
        self.write_output(self.param_kwargs)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.manifest import manifest_mixin
from servicecatalog_puppet.workflow.workspaces import provision_dry_run_workspace_task
from servicecatalog_puppet.workflow.workspaces import provision_workspace_task
from servicecatalog_puppet.workflow.workspaces import terminate_dry_run_workspace_task
from servicecatalog_puppet.workflow.workspaces import terminate_workspace_task
from servicecatalog_puppet.workflow.workspaces import workspace_base_task


class WorkspaceForTask(
    workspace_base_task.WorkspaceBaseTask, manifest_mixin.ManifestMixen
):
    workspace_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "workspace_name": self.workspace_name,
            "cache_invalidator": self.cache_invalidator,
        }

    def get_klass_for_provisioning(self):
        if self.is_dry_run:
            if self.status == constants.PROVISIONED:
                return provision_dry_run_workspace_task.ProvisionDryRunWorkspaceTask
            elif self.status == constants.TERMINATED:
                return terminate_dry_run_workspace_task.TerminateDryRunWorkspaceTask
            else:
                raise Exception(f"Unknown status: {self.status}")

        else:
            if self.status == constants.PROVISIONED:
                return provision_workspace_task.ProvisionWorkspaceTask
            elif self.status == constants.TERMINATED:
                return terminate_workspace_task.TerminateWorkspaceTask
            else:
                raise Exception(f"Unknown status: {self.status}")

    def run(self):
        self.write_output(self.params_for_results_display())



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import time

import luigi

from servicecatalog_puppet.workflow.launch import provisioning_task


class DoDescribeProvisioningParameters(provisioning_task.ProvisioningTask):
    puppet_account_id = luigi.Parameter()
    region = luigi.Parameter()
    product_id = luigi.Parameter()
    version_id = luigi.Parameter()
    portfolio = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "product_id": self.product_id,
            "version_id": self.version_id,
        }

    def api_calls_used(self):
        return [
            f"servicecatalog.describe_provisioning_parameters_{self.puppet_account_id}_{self.region}",
        ]

    def run(self):
        with self.hub_regional_client("servicecatalog") as service_catalog:

            provisioning_artifact_parameters = None
            retries = 3
            while retries > 0:
                try:
                    provisioning_artifact_parameters = service_catalog.describe_provisioning_parameters(
                        ProductId=self.product_id,
                        ProvisioningArtifactId=self.version_id,
                        PathName=self.portfolio,
                    ).get(
                        "ProvisioningArtifactParameters", []
                    )
                    retries = 0
                    break
                except service_catalog.exceptions.ClientError as ex:
                    if "S3 error: Access Denied" in str(ex):
                        self.info("Swallowing S3 error: Access Denied")
                    else:
                        raise ex
                    time.sleep(3)
                    retries -= 1

            self.write_output(
                provisioning_artifact_parameters
                if isinstance(provisioning_artifact_parameters, list)
                else [provisioning_artifact_parameters]
            )



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class DisassociateProductFromPortfolioTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    manifest_file_path = "manifest_file_path"
    account_id = "account_id"
    region = "region"
    portfolio_id = "portfolio_id"
    product_id = "product_id"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
            disassociate_product_from_portfolio_task,
        )

        self.module = disassociate_product_from_portfolio_task

        self.sut = self.module.DisassociateProductFromPortfolio(
            manifest_file_path=self.manifest_file_path,
            account_id=self.account_id,
            region=self.region,
            portfolio_id=self.portfolio_id,
            product_id=self.product_id,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "account_id": self.account_id,
            "region": self.region,
            "portfolio_id": self.portfolio_id,
            "product_id": self.product_id,
            "cache_invalidator": self.cache_invalidator,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_api_calls_used(self):
        # setup
        expected_result = {
            f"servicecatalog.disassociate_product_from_portfolio_{self.account_id}_{self.region}_{self.portfolio_id}_{self.product_id}": 1,
        }

        # exercise
        actual_result = self.sut.api_calls_used()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_run(self):
        # setup
        result = dict(foo="bar")
        self.inject_spoke_regional_client_called_with_response(
            "servicecatalog", "disassociate_product_from_portfolio", result
        )

        # exercise
        self.sut.run()

        # verify
        self.assert_spoke_regional_client_called_with(
            "servicecatalog",
            "disassociate_product_from_portfolio",
            dict(PortfolioId=self.portfolio_id, ProductId=self.product_id,),
        )
        self.assert_output(result)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import os
import shutil
from datetime import datetime
from pathlib import Path

import cfn_tools
import click
import terminaltables
import yaml
from betterboto import client as betterboto_client
from jinja2 import Template

from servicecatalog_puppet import constants, config, asset_helpers


def upload_config(config):
    with betterboto_client.ClientContextManager("ssm") as ssm:
        ssm.put_parameter(
            Name=constants.CONFIG_PARAM_NAME,
            Type="String",
            Value=yaml.safe_dump(config),
            Overwrite=True,
        )
    click.echo("Uploaded config")


def seed(complexity, p):
    example = "manifest-{}.yaml".format(complexity)
    shutil.copy2(
        asset_helpers.resolve_from_site_packages(
            os.path.sep.join(["manifests", example])
        ),
        os.path.sep.join([p, "manifest.yaml"]),
    )


def list_resources():
    click.echo("# Framework resources")

    click.echo("## SSM Parameters used")
    click.echo(f"- {constants.CONFIG_PARAM_NAME}")
    click.echo(f"- {constants.CONFIG_PARAM_NAME_ORG_IAM_ROLE_ARN}")

    for file in Path(__file__).parent.resolve().glob("*.template.yaml"):
        if "empty.template.yaml" == file.name:
            continue
        template_contents = Template(open(file, "r").read()).render()
        template = cfn_tools.load_yaml(template_contents)
        click.echo(f"## Resources for stack: {file.name.split('.')[0]}")
        table_data = [
            ["Logical Name", "Resource Type", "Name",],
        ]
        table = terminaltables.AsciiTable(table_data)
        for logical_name, resource in template.get("Resources").items():
            resource_type = resource.get("Type")
            name = "-"
            type_to_name = {
                "AWS::IAM::Role": "RoleName",
                "AWS::SSM::Parameter": "Name",
                "AWS::S3::Bucket": "BucketName",
                "AWS::CodePipeline::Pipeline": "Name",
                "AWS::CodeBuild::Project": "Name",
                "AWS::CodeCommit::Repository": "RepositoryName",
                "AWS::SNS::Topic": "TopicName",
                "AWS::SQS::Queue": "QueueName",
            }

            if type_to_name.get(resource_type) is not None:
                name = resource.get("Properties", {}).get(
                    type_to_name.get(resource_type), "Not Specified"
                )
                if not isinstance(name, str):
                    name = cfn_tools.dump_yaml(name)

            table_data.append([logical_name, resource_type, name])

        click.echo(table.table)
    click.echo(f"n.b. AWS::StackName evaluates to {constants.BOOTSTRAP_STACK_NAME}")


def set_config_value(name, value):
    with betterboto_client.ClientContextManager(
        "ssm", region_name=constants.HOME_REGION
    ) as ssm:
        try:
            response = ssm.get_parameter(Name=constants.CONFIG_PARAM_NAME)
            config = yaml.safe_load(response.get("Parameter").get("Value"))
        except ssm.exceptions.ParameterNotFound:
            config = {}

        if name == "regions":
            config["regions"] = value if len(value) > 1 else value[0].split(",")
        else:
            config[name] = value.upper() == "TRUE"

        upload_config(config)


def set_named_config_value(name, value):
    with betterboto_client.ClientContextManager(
        "ssm", region_name=constants.HOME_REGION
    ) as ssm:
        ssm.put_parameter(
            Name=name, Type="String", Value=value, Overwrite=True,
        )
        click.echo("Uploaded named config")


def handle_action_execution_detail(puppet_account_id, action_execution_detail):
    action_type_id = action_execution_detail.get("input").get("actionTypeId")
    if (
        action_type_id.get("category") == "Build"
        and action_type_id.get("owner") == "AWS"
        and action_type_id.get("provider") == "CodeBuild"
    ):
        external_execution_id = (
            action_execution_detail.get("output")
            .get("executionResult")
            .get("externalExecutionId")
        )

        with betterboto_client.ClientContextManager(
            "codebuild", region_name=config.get_home_region(puppet_account_id)
        ) as codebuild:
            builds = codebuild.batch_get_builds(ids=[external_execution_id]).get(
                "builds"
            )
            build = builds[0]
            log_details = build.get("logs")
            with betterboto_client.ClientContextManager(
                "logs", region_name=config.get_home_region(puppet_account_id)
            ) as logs:
                with open(
                    f"log-{action_execution_detail.get('input').get('configuration').get('ProjectName')}.log",
                    "w",
                ) as f:
                    params = {
                        "logGroupName": log_details.get("groupName"),
                        "logStreamName": log_details.get("streamName"),
                        "startFromHead": True,
                    }
                    has_more_logs = True
                    while has_more_logs:
                        get_log_events_response = logs.get_log_events(**params)
                        if (len(get_log_events_response.get("events"))) > 0:
                            params["nextToken"] = get_log_events_response.get(
                                "nextForwardToken"
                            )
                        else:
                            has_more_logs = False
                            if params.get("nextToken"):
                                del params["nextToken"]
                        for e in get_log_events_response.get("events"):
                            d = datetime.utcfromtimestamp(
                                e.get("timestamp") / 1000
                            ).strftime("%Y-%m-%d %H:%M:%S")
                            f.write(f"{d} : {e.get('message')}")


def export_puppet_pipeline_logs(execution_id, puppet_account_id):
    with betterboto_client.ClientContextManager(
        "codepipeline", region_name=config.get_home_region(puppet_account_id)
    ) as codepipeline:
        action_execution_details = codepipeline.list_action_executions(
            pipelineName=constants.PIPELINE_NAME,
            filter={"pipelineExecutionId": execution_id},
        ).get("actionExecutionDetails")

        for action_execution_detail in action_execution_details:
            handle_action_execution_detail(puppet_account_id, action_execution_detail)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import config
from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.manifest import manifest_mixin
from servicecatalog_puppet.workflow.portfolio.accessors import (
    get_portfolio_by_portfolio_name_task,
)
from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
    portfolio_management_task,
)
from servicecatalog_puppet.workflow.portfolio.sharing_management import (
    share_portfolio_task,
)
from servicecatalog_puppet.workflow.portfolio.sharing_management import (
    share_portfolio_via_orgs_task,
)


class ShareAndAcceptPortfolioTask(
    portfolio_management_task.PortfolioManagementTask, manifest_mixin.ManifestMixen
):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    puppet_account_id = luigi.Parameter()
    sharing_mode = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
            "sharing_mode": self.sharing_mode,
        }

    def requires(self):
        requirements = dict(
            portfolio=get_portfolio_by_portfolio_name_task.GetPortfolioByPortfolioName(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                portfolio=self.portfolio,
                account_id=self.puppet_account_id,
                region=self.region,
            )
        )
        if self.sharing_mode == constants.SHARING_MODE_AWS_ORGANIZATIONS:
            ou_to_share_with = self.manifest.get_account(self.account_id).get(
                "expanded_from"
            )
            if ou_to_share_with is None:
                self.warning(
                    f"Sharing {self.portfolio} with {self.account_id} not using orgs as there is no OU set for it"
                )
                requirements["share"] = share_portfolio_task.SharePortfolioTask(
                    manifest_file_path=self.manifest_file_path,
                    portfolio=self.portfolio,
                    account_id=self.account_id,
                    region=self.region,
                    puppet_account_id=self.puppet_account_id,
                )
            else:
                requirements[
                    "share"
                ] = share_portfolio_via_orgs_task.SharePortfolioViaOrgsTask(
                    manifest_file_path=self.manifest_file_path,
                    portfolio=self.portfolio,
                    region=self.region,
                    puppet_account_id=self.puppet_account_id,
                    ou_to_share_with=ou_to_share_with,
                )

        else:
            requirements["share"] = share_portfolio_task.SharePortfolioTask(
                manifest_file_path=self.manifest_file_path,
                portfolio=self.portfolio,
                account_id=self.account_id,
                region=self.region,
                puppet_account_id=self.puppet_account_id,
            )
        return requirements

    def api_calls_used(self):
        return [
            f"servicecatalog.list_accepted_portfolio_shares_single_page{self.account_id}_{self.region}",
            f"servicecatalog.accept_portfolio_share{self.account_id}_{self.region}",
            f"servicecatalog.associate_principal_with_portfolio{self.account_id}_{self.region}",
            f"servicecatalog.list_principals_for_portfolio{self.account_id}_{self.region}",
        ]

    def run(self):
        portfolio = self.load_from_input("portfolio")
        portfolio_id = portfolio.get("portfolio_id")

        with self.spoke_regional_client(
            "servicecatalog"
        ) as cross_account_servicecatalog:
            if self.sharing_mode != constants.SHARING_MODE_AWS_ORGANIZATIONS:
                was_accepted = False
                accepted_portfolio_shares = cross_account_servicecatalog.list_accepted_portfolio_shares_single_page().get(
                    "PortfolioDetails"
                )
                for accepted_portfolio_share in accepted_portfolio_shares:
                    if accepted_portfolio_share.get("Id") == portfolio_id:
                        was_accepted = True
                        break
                if not was_accepted:
                    self.info(f"{self.uid}: accepting {portfolio_id}")
                    cross_account_servicecatalog.accept_portfolio_share(
                        PortfolioId=portfolio_id
                    )

            principals_for_portfolio = cross_account_servicecatalog.list_principals_for_portfolio_single_page(
                PortfolioId=portfolio_id
            ).get(
                "Principals"
            )
            puppet_role_needs_associating = True
            principal_to_associate = config.get_puppet_role_arn(self.account_id)
            self.info(f"Checking if {principal_to_associate} needs to be associated")
            for principal_for_portfolio in principals_for_portfolio:
                self.info(
                    f"comparing {principal_to_associate} to {principal_for_portfolio.get('PrincipalARN')}"
                )
                if (
                    principal_for_portfolio.get("PrincipalARN")
                    == principal_to_associate
                ):
                    self.info("found a match!! no longer going to add")
                    puppet_role_needs_associating = False

            if puppet_role_needs_associating:
                cross_account_servicecatalog.associate_principal_with_portfolio(
                    PortfolioId=portfolio_id,
                    PrincipalARN=principal_to_associate,
                    PrincipalType="IAM",
                )

        self.write_output(self.param_kwargs)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class AssertionForAccountAndRegionTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    manifest_file_path = "manifest_file_path"
    account_id = "account_id"
    region = "region"
    assertion_name = "assertion_name"
    puppet_account_id = "puppet_account_id"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.assertions import (
            assertion_for_account_and_region_task,
        )

        self.module = assertion_for_account_and_region_task

        self.sut = self.module.AssertionForAccountAndRegionTask(
            puppet_account_id=self.puppet_account_id,
            assertion_name=self.assertion_name,
            manifest_file_path=self.manifest_file_path,
            account_id=self.account_id,
            region=self.region,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "assertion_name": self.assertion_name,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class EnsureEventBridgeEventBusTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    puppet_account_id = "puppet_account_id"
    region = "region"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.generate import (
            ensure_event_bridge_event_bus_task,
        )

        self.module = ensure_event_bridge_event_bus_task

        self.sut = self.module.EnsureEventBridgeEventBusTask(
            puppet_account_id=self.puppet_account_id, region=self.region
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_api_calls_used(self):
        # setup
        expected_result = {
            f"events.describe_event_bus_{self.puppet_account_id}_{self.region}": 1,
            f"events.create_event_bus_{self.puppet_account_id}_{self.region}": 1,
        }

        # exercise
        actual_result = self.sut.api_calls_used()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_run(self):
        # setup
        from servicecatalog_puppet import constants

        # exercise
        self.sut.run()

        # verify
        self.assert_hub_regional_client_called_with(
            "events", "describe_event_bus", dict(Name=constants.EVENT_BUS_NAME)
        )
        self.assert_output(True)



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
    portfolio_management_task,
)


class GetPortfolioByPortfolioName(portfolio_management_task.PortfolioManagementTask):
    puppet_account_id = luigi.Parameter()
    portfolio = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def api_calls_used(self):
        return [
            f"servicecatalog.list_accepted_portfolio_shares_single_page{self.account_id}_{self.region}",
            f"servicecatalog.list_portfolios_{self.account_id}_{self.region}",
        ]

    def run(self):
        with self.spoke_regional_client(
            "servicecatalog"
        ) as cross_account_servicecatalog:
            result = None
            response = cross_account_servicecatalog.list_accepted_portfolio_shares_single_page(
                PortfolioShareType="AWS_ORGANIZATIONS"
            )
            for portfolio_detail in response.get("PortfolioDetails"):
                if portfolio_detail.get("DisplayName") == self.portfolio:
                    result = portfolio_detail
                    break

            if result is None:
                response = cross_account_servicecatalog.list_portfolios_single_page()
                for portfolio_detail in response.get("PortfolioDetails", []):
                    if portfolio_detail.get("DisplayName") == self.portfolio:
                        result = portfolio_detail
                        break

            assert result is not None, "Could not find portfolio"
        self.write_output(
            {
                "portfolio_name": self.portfolio,
                "portfolio_id": result.get("Id"),
                "provider_name": result.get("ProviderName"),
                "description": result.get("Description"),
            }
        )



#  Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.launch import run_deploy_in_spoke_task
from servicecatalog_puppet.workflow.manifest import section_task


class GenericSectionTask(section_task.SectionTask):

    section_name_singular = "not_set"
    section_name = "not_set"
    for_region_task_klass = "not_set"
    for_account_task_klass = "not_set"
    for_account_and_region_task_klass = "not_set"
    task_klass = "not_set"
    item_name = "not_set"
    supports_spoke_mode = False

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        requirements = list()
        common_args = dict(
            puppet_account_id=self.puppet_account_id,
            manifest_file_path=self.manifest_file_path,
        )

        for name, details in self.manifest.get(self.section_name, {}).items():
            if (
                details.get(constants.MANIFEST_STATUS_FIELD_NAME)
                != constants.MANIFEST_STATUS_FIELD_VALUE_IGNORED
            ):
                common_args[self.item_name] = name
                requirements += self.handle_requirements_for(
                    name,
                    self.section_name_singular,
                    self.section_name,
                    self.for_region_task_klass,
                    self.for_account_task_klass,
                    self.for_account_and_region_task_klass,
                    self.task_klass,
                    common_args,
                    self.supports_spoke_mode,
                )

        return requirements

    def run(self):
        if self.supports_spoke_mode and not self.is_running_in_spoke():
            tasks_to_run = list()
            for name, details in self.manifest.get(self.section_name, {}).items():
                if (
                    details.get(constants.MANIFEST_STATUS_FIELD_NAME)
                    != constants.MANIFEST_STATUS_FIELD_VALUE_IGNORED
                ):
                    if (
                        details.get("execution", constants.EXECUTION_MODE_DEFAULT)
                        == constants.EXECUTION_MODE_SPOKE
                    ):
                        for (
                            account_id
                        ) in self.manifest.get_account_ids_used_for_section_item(
                            self.puppet_account_id, self.section_name, name
                        ):
                            tasks_to_run.append(
                                run_deploy_in_spoke_task.RunDeployInSpokeTask(
                                    manifest_file_path=self.manifest_file_path,
                                    puppet_account_id=self.puppet_account_id,
                                    account_id=account_id,
                                )
                            )
            yield tasks_to_run
        self.write_output(self.manifest.get(self.section_name, {}))
